import torch
import torch.nn as nn
from torch_geometric.data import Data, Batch
import torch.nn.functional as F
import numpy as np
from torch_geometric.utils import degree
from ..base import BaseRandom

def split_batch(g, edge_index):
    split = degree(g.batch[edge_index[0]], dtype=torch.long).tolist()
    edge_indices = torch.split(edge_index, split, dim=1)
    num_nodes = degree(g.batch, dtype=torch.long)
    cum_nodes = torch.cat([g.batch.new_zeros(1), num_nodes.cumsum(dim=0)[:-1]])
    num_edges = torch.tensor([e.size(1) for e in edge_indices], dtype=torch.long).to(g.x.device)
    cum_edges = torch.cat([g.batch.new_zeros(1), num_edges.cumsum(dim=0)[:-1]])

    return edge_indices, num_nodes, cum_nodes, num_edges, cum_edges

def relabel(x, edge_index, batch, pos=None):

    num_nodes = x.size(0)
    sub_nodes = torch.unique(edge_index)
    x = x[sub_nodes]
    # try:
    #     assert batch[sub_nodes].unique().shape[0] == 128
    batch = batch[sub_nodes]

    row, col = edge_index
    # remapping the nodes in the explanatory subgraph to new ids.
    node_idx = row.new_full((num_nodes,), -1)
    node_idx[sub_nodes] = torch.arange(sub_nodes.size(0), device=x.device)
    edge_index = node_idx[edge_index]
    if pos is not None:
        pos = pos[sub_nodes]
    return x, edge_index, batch, pos


class CIGA(BaseRandom):

    def __init__(self, clf, extractor, criterion, config):
        super().__init__()
        self.name = 'ciga'
        self.clf = clf
        self.extractor = extractor
        self.criterion = criterion
        self.device = next(self.parameters()).device

        self.contrast_loss_coef = config['contrast_loss_coef']
        self.hinge_loss_coef = config['hinge_loss_coef']
        # self.reg_loss_coef = config.reg_loss_coef
        self.causal_ratio = config['causal_ratio']
        # self.contrast_method = config['contrast_method']



    def __loss__(self, causal_logits, conf_logits, clf_labels, contrast_loss):
        # pred_loss = self.criterion()
        pred_loss = self.criterion(causal_logits, clf_labels.float())
        causal_loss = self.criterion(causal_logits, clf_labels.float(), reduction='none').reshape(-1)
        conf_loss = self.criterion(conf_logits, clf_labels.float(), reduction='none').reshape(-1)
        # causal_loss = [self.criterion(causal_logits[i], clf_labels[i].float()) for i in range(conf_logits.shape[0])]
        # conf_loss = [self.criterion(conf_logits[i], clf_labels[i].float()) for i in range(conf_logits.shape[0])]

        spu_loss_weight = torch.zeros(conf_loss.size()).to(clf_labels.device)
        spu_loss_weight[conf_loss > causal_loss] = 1.0
        hinge_loss = conf_loss.dot(spu_loss_weight) / ((conf_loss > causal_loss).sum() + 1e-6)

        hinge_loss = self.hinge_loss_coef * hinge_loss
        contrast_loss = self.contrast_loss_coef * contrast_loss

        loss = pred_loss + hinge_loss + contrast_loss
        loss_dict = {'loss': loss.item(), 'pred': pred_loss.item(), 'contrast': contrast_loss.item(), 'hinge': hinge_loss.item()}
        return loss, loss_dict

    def forward(self, data, edge_attn=None):
        emb, edge_index = self.clf.get_emb(data, edge_attn=edge_attn)
        node_weight = self.extractor(emb, batch=data.batch, pool_out_lig=None)
        pred_edge_weight = self.node_attn_to_edge_attn(node_weight, edge_index)
        # pred_edge_weight = self.linear(edge_rep).view(-1)
        pred_edge_weight = pred_edge_weight.view(-1)

        causal_edge_index = torch.LongTensor([[], []]).to(self.device)
        causal_edge_weight = torch.tensor([]).to(self.device)

        edge_indices, _, _, num_edges, cum_edges = split_batch(data, edge_index)
        for edge_index, N, C in zip(edge_indices, num_edges, cum_edges):
            n_reserve = int(self.causal_ratio * N)
            single_mask = pred_edge_weight[C:C + N]
            single_mask_detach = pred_edge_weight[C:C + N].detach()
            rank = np.argpartition(-single_mask_detach.cpu().numpy(), n_reserve)
            idx_reserve, idx_drop = rank[:n_reserve], rank[n_reserve:]
            # the edges are quite a few so that we directly use the original graph
            if n_reserve == 0:
                causal_edge_index = torch.cat([causal_edge_index, edge_index], dim=1)
                causal_edge_weight = torch.cat([causal_edge_weight, single_mask])
            else:
                causal_edge_index = torch.cat([causal_edge_index, edge_index[:, idx_reserve]], dim=1)
                causal_edge_weight = torch.cat([causal_edge_weight, single_mask[idx_reserve]])

        causal_x, causal_edge_index, causal_batch, causal_pos = relabel(data.x, causal_edge_index, data.batch, data.pos)

        if hasattr(data, "x_lig"):
            causal_data = Batch(batch=causal_batch, edge_index=causal_edge_index, x=causal_x, pos=causal_pos,
                            x_lig=data.x_lig, pos_lig=data.pos_lig, x_lig_batch=data.x_lig_batch)
        else:
            causal_data = Batch(batch=causal_batch, edge_index=causal_edge_index, x=causal_x, pos=causal_pos)

        causal_node_emb = self.clf.get_emb(causal_data, edge_attn=causal_data.edge_weight)[0]
        causal_logits = self.clf.get_pred_from_emb(causal_node_emb, causal_data.batch)
        return causal_logits

    def forward_pass(self, data, epoch, do_sampling, **kwargs):
        # x = self.clf.get_emb(data, edge_attr=data.edge_attr)
        emb, edge_index = self.clf.get_emb(data)
        # original_clf_logits = self.clf(data, edge_attr=data.edge_attr)
        node_weight = self.extractor(emb, batch=data.batch, pool_out_lig=None)
        pred_edge_weight = self.node_attn_to_edge_attn(node_weight, edge_index)
        # pred_edge_weight = self.linear(edge_rep).view(-1)

        pred_edge_weight = pred_edge_weight.view(-1)
        device = data.x.device
        causal_edge_index = torch.LongTensor([[], []]).to(self.device)
        causal_edge_weight = torch.tensor([]).to(self.device)
        # causal_edge_attr = torch.tensor([]).to(device)
        spu_edge_index = torch.LongTensor([[], []]).to(self.device)
        spu_edge_weight = torch.tensor([]).to(self.device)

        edge_indices, _, _, num_edges, cum_edges = split_batch(data, edge_index)
        for edge_index, N, C in zip(edge_indices, num_edges, cum_edges):
            n_reserve = int(self.causal_ratio * N)

            single_mask = pred_edge_weight[C:C + N]
            single_mask_detach = pred_edge_weight[C:C + N].detach()
            rank = np.argpartition(-single_mask_detach.cpu().numpy(), n_reserve)
            idx_reserve, idx_drop = rank[:n_reserve], rank[n_reserve:]

            causal_edge_index = torch.cat([causal_edge_index, edge_index[:, idx_reserve]], dim=1)
            spu_edge_index = torch.cat([spu_edge_index, edge_index[:, idx_drop]], dim=1)

            causal_edge_weight = torch.cat([causal_edge_weight, single_mask[idx_reserve]])
            spu_edge_weight = torch.cat([spu_edge_weight, -1 * single_mask[idx_drop]])

            # causal_edge_attr = torch.cat([causal_edge_attr, edge_attr[idx_reserve]])
            # spu_edge_attr = torch.cat([spu_edge_attr, edge_attr[idx_drop]])

        causal_x, causal_edge_index, causal_batch, causal_pos = relabel(data.x, causal_edge_index, data.batch, data.pos)
        spu_x, spu_edge_index, spu_batch, spu_pos = relabel(data.x, spu_edge_index, data.batch, data.pos)

        if hasattr(data, "x_lig"):
            causal_data = Batch(batch=causal_batch, edge_index=causal_edge_index, x=causal_x, pos=causal_pos,
                            x_lig=data.x_lig, pos_lig=data.pos_lig, x_lig_batch=data.x_lig_batch)
            conf_data = Batch(batch=spu_batch, edge_index=spu_edge_index, x=spu_x, pos=spu_pos,
                              x_lig=data.x_lig, pos_lig=data.pos_lig, x_lig_batch=data.x_lig_batch)
        else:
            causal_data = Batch(batch=causal_batch, edge_index=causal_edge_index, x=causal_x, pos=causal_pos)
            conf_data = Batch(batch=spu_batch, edge_index=spu_edge_index, x=spu_x, pos=spu_pos)

        causal_node_emb = self.clf.get_emb(causal_data, edge_attn=causal_data.edge_weight)[0]
        conf_node_emb = self.clf.get_emb(conf_data, edge_attn=conf_data.edge_weight)[0]
        causal_graph_emb = self.clf.pool(causal_node_emb, causal_data.batch)

        causal_logits = self.clf.get_pred_from_emb(causal_node_emb, causal_data.batch)
        # this is not the same as the origin dir
        conf_logits = self.clf.get_pred_from_emb(conf_node_emb, conf_data.batch)


        contrast_loss = self.get_contrast_loss(causal_graph_emb, data.y.reshape(-1), norm=1)
        loss, loss_dict = self.__loss__(causal_logits, conf_logits, data.y, contrast_loss)

        res_weights = self.node_attn_to_edge_attn(node_weight, data.edge_index) if hasattr(data, 'edge_label') else node_weight

        return loss, loss_dict, causal_logits, res_weights.reshape(-1)


    def create_new_data(self, graph, idx, edge_weight):
        x = graph.x
        edge_index = graph.edge_index[:, idx]
        edge_attr = graph.edge_attr[idx] if graph.edge_attr is not None else None

        # node relabel
        num_nodes = x.size(0)
        sub_nodes = torch.unique(edge_index)

        x = x[sub_nodes]

        pos = graph.pos[sub_nodes] if graph.pos is not None else None
        # pos = pos

        row, col = edge_index
        # remapping the nodes in the explanatory subgraph to new ids.
        node_idx = row.new_full((num_nodes,), -1)
        node_idx[sub_nodes] = torch.arange(sub_nodes.size(0), device=row.device)
        edge_index = node_idx[edge_index]

        return Data(x=x, edge_index=edge_index, pos=pos, edge_attr=edge_attr, edge_weight=edge_weight)

    def get_contrast_loss(self, causal_rep, labels, norm=None, contrast_t=1.0, sampling='mul', y_pred=None):

        if norm != None:
            causal_rep = F.normalize(causal_rep)
        if sampling.lower() in ['mul', 'var']:
            # modified from https://github.com/HobbitLong/SupContrast/blob/master/losses.py#L11
            device = causal_rep.device
            mask = torch.eq(labels.unsqueeze(1), labels.unsqueeze(1).T).float().to(device)
            # compute logits
            anchor_dot_contrast = torch.div(torch.matmul(causal_rep, causal_rep.T), contrast_t)
            # for numerical stability
            logits_max, _ = torch.max(anchor_dot_contrast, dim=1, keepdim=True)
            logits = anchor_dot_contrast - logits_max.detach()
            # tile mask: no need
            # mask = mask.repeat(anchor_count, contrast_count)
            batch_size = labels.size(0)
            anchor_count = 1
            # mask-out self-contrast cases (set the diagonal to 0)
            logits_mask = torch.scatter(torch.ones_like(mask), 1,
                                        torch.arange(batch_size * anchor_count).view(-1, 1).to(device), 0)
            mask = mask * logits_mask
            # compute log_prob
            exp_logits = torch.exp(logits) * logits_mask
            log_prob = logits - torch.log(exp_logits.sum(1, keepdim=True))
            # min_value = torch.min(log_prob)
            # log_prob = torch.where(log_prob == torch.inf, 2 * min_value, log_prob)

            # print(log_prob)
            # print(mask.sum(1))
            # compute mean of log-likelihood over positive
            is_valid = mask.sum(1) != 0
            mean_log_prob_pos = (mask * log_prob).sum(1)[is_valid] / mask.sum(1)[is_valid]
            # some classes may not be sampled by more than 2
            # mean_log_prob_pos[torch.isnan(mean_log_prob_pos)] = 0.0

            # loss
            # contrast_loss = -(args.temperature / args.base_temperature) * mean_log_prob_pos
            # contrast_loss = contrast_loss.view(anchor_count, batch_size).mean()
            contrast_loss = -mean_log_prob_pos.mean()
            if sampling.lower() == 'var':
                contrast_loss += mean_log_prob_pos.var()
        elif sampling.lower() == 'single':
            N = causal_rep.size(0)
            pos_idx = torch.arange(N)
            neg_idx = torch.randperm(N)
            for i in range(N):
                for j in range(N):
                    if labels[i] == labels[j]:
                        pos_idx[i] = j
                    else:
                        neg_idx[i] = j
            contrast_loss = -torch.mean(
                torch.bmm(causal_rep.unsqueeze(1), causal_rep[pos_idx].unsqueeze(1).transpose(1, 2)) -
                torch.matmul(causal_rep.unsqueeze(1), causal_rep[neg_idx].unsqueeze(1).transpose(1, 2)))
        elif sampling.lower() == 'cncp':
            # correct & contrast with hard postive only https://arxiv.org/abs/2203.01517
            device = causal_rep.device
            mask = torch.eq(labels.unsqueeze(1), labels.unsqueeze(1).T).float().to(device)
            # compute logits
            anchor_dot_contrast = torch.div(torch.matmul(causal_rep, causal_rep.T), contrast_t)
            # for numerical stability
            logits_max, _ = torch.max(anchor_dot_contrast, dim=1, keepdim=True)
            logits = anchor_dot_contrast - logits_max.detach()
            # tile mask: no need
            # mask = mask.repeat(anchor_count, contrast_count)
            batch_size = labels.size(0)
            anchor_count = 1
            # mask-out self-contrast cases
            logits_mask = torch.scatter(torch.ones_like(mask), 1,
                                        torch.arange(batch_size * anchor_count).view(-1, 1).to(device), 0)
            mask = mask * logits_mask
            # find hard postive & negative
            pos_mask = y_pred != labels
            neg_mask = y_pred == labels

            # hard negative: diff label && correct pred
            neg_mask = torch.logical_not(mask)  # * neg_mask
            # hard positive: same label && incorrect pred
            pos_mask = mask * pos_mask

            # compute log_prob
            neg_exp_logits = torch.exp(logits) * neg_mask
            pos_exp_logits = torch.exp(logits) * pos_mask
            log_prob = logits - \
                       torch.log(pos_exp_logits.sum(1, keepdim=True) + \
                                 neg_exp_logits.sum(1, keepdim=True))

            # compute mean of log-likelihood over positive
            is_valid = pos_mask.sum(1) != 0
            mean_log_prob_pos = (pos_mask * log_prob).sum(1)[is_valid] / pos_mask.sum(1)[is_valid]
            # some classes may not be sampled by more than 2
            # mean_log_prob_pos[torch.isnan(mean_log_prob_pos)] = 0.0

            # loss
            # contrast_loss = -(args.temperature / args.base_temperature) * mean_log_prob_pos
            # contrast_loss = contrast_loss.view(anchor_count, batch_size).mean()
            contrast_loss = -mean_log_prob_pos.mean()
        elif sampling.lower() == 'cnc':
            # correct & contrast https://arxiv.org/abs/2203.01517
            device = causal_rep.device
            mask = torch.eq(labels.unsqueeze(1), labels.unsqueeze(1).T).float().to(device)
            # compute logits
            anchor_dot_contrast = torch.div(torch.matmul(causal_rep, causal_rep.T), contrast_t)
            # for numerical stability
            logits_max, _ = torch.max(anchor_dot_contrast, dim=1, keepdim=True)
            logits = anchor_dot_contrast - logits_max.detach()
            # tile mask: no need
            # mask = mask.repeat(anchor_count, contrast_count)
            batch_size = labels.size(0)
            anchor_count = 1
            # mask-out self-contrast cases
            logits_mask = torch.scatter(torch.ones_like(mask), 1,
                                        torch.arange(batch_size * anchor_count).view(-1, 1).to(device), 0)
            mask = mask * logits_mask
            # find hard postive & negative
            pos_mask = y_pred != labels
            neg_mask = y_pred == labels
            # hard negative: diff label && correct pred
            neg_mask = torch.logical_not(mask) * neg_mask * logits_mask
            # hard positive: same label && incorrect pred
            pos_mask = mask * pos_mask
            if neg_mask.sum() == 0:
                neg_mask = torch.logical_not(mask)
            if pos_mask.sum() == 0:
                pos_mask = mask
            # compute log_prob
            neg_exp_logits = torch.exp(logits) * neg_mask
            pos_exp_logits = torch.exp(logits) * pos_mask
            log_prob = logits - \
                       torch.log(pos_exp_logits.sum(1, keepdim=True) + \
                                 neg_exp_logits.sum(1, keepdim=True) + 1e-12)
            # compute mean of log-likelihood over positive
            is_valid = pos_mask.sum(1) != 0
            mean_log_prob_pos = (pos_mask * log_prob).sum(1)[is_valid] / pos_mask.sum(1)[is_valid]
            contrast_loss = -mean_log_prob_pos.mean()
        else:
            raise Exception("Not implmented contrasting method")
        return contrast_loss

    @staticmethod
    def node_attn_to_edge_attn(node_attn, edge_index):
        src_attn = node_attn[edge_index[0]]
        dst_attn = node_attn[edge_index[1]]
        # edge_attn = 1 - (1 - src_attn) * (1 - dst_attn)
        edge_attn = src_attn * dst_attn
        # edge_attn = (src_attn + dst_attn) / 2
        return edge_attn



import torch
import torch.nn as nn
from torch_scatter import scatter
import numpy as np
from ..base import BaseRandom
from torch_geometric.nn import ASAPooling as BasePooling
import torch.nn.functional as F
from torch_scatter import scatter
from torch_sparse import SparseTensor
from torch_geometric.nn.pool.topk_pool import topk
from torch_geometric.utils import add_remaining_self_loops, softmax


class ASAPooling(BasePooling):
    def __init__(self, *args, **kargs):
        super(ASAPooling, self).__init__(*args, **kargs)

    def forward(self, x, edge_index, edge_weight=None, batch=None, return_attn=None):
        """"""
        N = x.size(0)

        edge_index, edge_weight = add_remaining_self_loops(
            edge_index, edge_weight, fill_value=1., num_nodes=N)

        if batch is None:
            batch = edge_index.new_zeros(x.size(0))

        x = x.unsqueeze(-1) if x.dim() == 1 else x

        x_pool = x
        if self.GNN is not None:
            x_pool = self.gnn_intra_cluster(x=x, edge_index=edge_index,
                                            edge_weight=edge_weight)

        x_pool_j = x_pool[edge_index[0]]
        x_q = scatter(x_pool_j, edge_index[1], dim=0, reduce='max')
        x_q = self.lin(x_q)[edge_index[1]]

        score = self.att(torch.cat([x_q, x_pool_j], dim=-1)).view(-1)
        score = F.leaky_relu(score, self.negative_slope)
        score = softmax(score, edge_index[1], num_nodes=N)

        # Sample attention coefficients stochastically.
        score = F.dropout(score, p=self.dropout, training=self.training)

        v_j = x[edge_index[0]] * score.view(-1, 1)
        x = scatter(v_j, edge_index[1], dim=0, reduce='add')

        # Cluster selection.
        fitness = self.gnn_score(x, edge_index).sigmoid().view(-1)
        perm = topk(fitness, self.ratio, batch)
        x = x[perm] * fitness[perm].view(-1, 1)
        batch = batch[perm]

        # Graph coarsening.
        row, col = edge_index
        A = SparseTensor(row=row, col=col, value=edge_weight,
                         sparse_sizes=(N, N))
        S = SparseTensor(row=row, col=col, value=score, sparse_sizes=(N, N))
        S = S[:, perm]

        A = S.t() @ A @ S

        if self.add_self_loops:
            A = A.fill_diag(1.)
        else:
            A = A.remove_diag()

        row, col, edge_weight = A.coo()
        edge_index = torch.stack([row, col], dim=0)
        if return_attn == 'fitness':
            return x, edge_index, edge_weight, batch, fitness
        else:
            assert return_attn == 'none'
            return x, edge_index, edge_weight, batch, perm


class ASAP(BaseRandom):

    def __init__(self, clf, extractor, criterion, config):
        super().__init__()

        self.name = 'asap'
        self.clf = clf
        self.extractor = extractor
        self.criterion = criterion
        self.device = next(self.parameters()).device

        self.return_attn = config['return_attn']

    def __loss__(self, clf_logits, clf_labels):
        pred_loss = self.criterion(clf_logits, clf_labels.float())
        loss_dict = {'loss': pred_loss.item(), 'pred': pred_loss.item()}
        return pred_loss, loss_dict

    def forward(self, data, edge_attn=None):
        emb, edge_index = self.clf.get_emb(data, edge_attn=edge_attn)
        # extractor is ASAPooling
        casual_emb, edge_index, causal_edge_weight, batch, perm = self.extractor(emb, edge_index, batch=data.batch, return_attn=self.return_attn)
        masked_clf_logits = self.clf.get_pred_from_emb(casual_emb, batch)
        return masked_clf_logits

    def forward_pass(self, data, epoch, do_sampling):

        emb, edge_index = self.clf.get_emb(data)
        # node_attn_log_logits = self.extractor(emb, batch=data.batch, pool_out_lig=None)
        casual_emb, edge_index, causal_edge_weight, batch, attn = self.extractor(emb, edge_index, batch=data.batch, return_attn=self.return_attn)

        masked_clf_logits = self.clf.get_pred_from_emb(casual_emb, batch)
        if self.return_attn == 'none':
            node_attn = torch.zeros(data.x.size(0), device=self.device)
            node_attn[attn] = 1
        else:
            node_attn = attn
        # node_attn = self.sampling(node_attn_log_logits, do_sampling)
        # node_attn = node_attn * (data.num_nodes / node_attn.sum())
        # edge_attn = self.node_attn_to_edge_attn(node_attn, edge_index)
        # masked_clf_logits = self.clf.get_pred_from_emb(node_attn * emb, batch=data.batch)
        # masked_clf_logits = self.clf(data, edge_attn=edge_attn)
        # original_clf_logits = self.clf(data)

        loss, loss_dict = self.__loss__(masked_clf_logits, data.y)
        return loss, loss_dict, masked_clf_logits, node_attn.reshape(-1)




from baselines import *

class DataUtil:
    def build():
        pass
    def get_train_dataloader():
        pass
    def get_val_dataloader():
        pass
    def get_test_dataloader():
        pass

from enum import Enum
class ModelUtil(Enum):
    model1 = ASAP
    model2 = LRIBern
    def build(self, args):
        cls_name = args.model_name
        self.model = self.cls_name.value(args)
        return self.model
    
    @staticmethod
    def save(model, metric, path):
        import torch
        if metric.best_perf.value > metric.compute():
            torch.save(model.state_dict(), path)


class OptimizerUtil:
    def build(self, args):
        return None

from collections import namedtuple

class Trainer:
    BestPerf = namedtuple("BestPerf", ['epoch', 'value'])
    def __init__(self, args) -> None:
        self.data = DataUtil.build(args)
        self.model = ModelUtil.build(args)
        self.optimizer = OptimizerUtil.build(args)
        self.metric = None
        self.loss_fn = self.optimizer.loss_fn

    def evaluate(self, dataset, model, optimizer):
        self.metric.reset()
        for batch in range(self.data):
            data, targets = batch
            self._eval_batch(data, targets)
        epoch_ret = self.metric.compute()
        return epoch_ret

    def train(self, val_dataset, model, trial=None, ckpt=None):
        info = print
        metric = self.metric
        for epoch in range(10):
            metric.reset()    
            for batch in range(self.data):
                data, target = batch
                self._train_batch(data, target) 
            epoch_ret = metric.compute()

            # compute best performance for epoch selection
            metric.best_perf = Trainer.BestPerf(epoch, max(metric.best_perf, epoch_ret))

            # report result
            info(epoch_ret, metric.best_perf)

            # save model
            ModelUtil.save(self.model, metric, path='')
        info("All interations done")

    def predict(self, **kwargs):
        self.evaluate(**kwargs)

    def _eval_batch(self, data, target):
        #shared by predict and evaluate
        preds = self.model(data) 
        ret = self.metric.update(preds, target)
        return ret

    def _train_batch(self, data, targets):
        self.optimizer.zero_grad() # reset grad
        preds = self.model(data)
        ret = self.metric.update(preds, targets) # update metric
        loss = self.loss_fn(preds, targets)
        loss.backward() # update grad
        self.optimizer.step() # compute grad


import torch
import torch.nn as nn
import numpy as np
import torch.nn.functional as F
from torch_scatter import scatter, scatter_max
from ...backbones import GINConv
# from evaluation import control_sparsity
from ..base import BaseRandom


class GNNLRP(BaseRandom):

    def __init__(self, clf, criterion, config):
        super().__init__()
        self.name = 'gnnlrp'
        self.clf = clf
        if hasattr(clf, 'clf'):
            self.target_layers = clf.clf.model.convs
        else:
            self.target_layers = clf.model.convs
        n_layers = len(self.target_layers)
        self.gammas = config.get("gammas", np.linspace(3, 0, n_layers-1))

        self.criterion = criterion
        self.device = next(self.parameters()).device

    def start_tracking(self):
        if hasattr(self.clf, 'clf'):
            self.activations_and_grads = Transforms(self.clf.clf, self.target_layers)
        else:
            self.activations_and_grads = Transforms(self.clf, self.target_layers)

    def _new_forward_pass(self, data, epoch, do_sampling, fp_16=True):
        original_clf_logits = self.activations_and_grads(data)
        H_end = self.activations_and_grads.activations[-1].data
        H_end = H_end.half() if fp_16 else H_end
        sum_nodes = 0
        node_weights_list = []
        for graph in data.to_data_list():
            add_nodes = graph.num_nodes
            relevance = H_end[sum_nodes:sum_nodes+add_nodes, :]
            # transforms = self.get_single_graph_transforms(graph, self.gammas, start=sum_nodes, fp_16=fp_16)
            generator = self.reverse_transforms_generator(graph, self.gammas, start=sum_nodes, fp_16=fp_16)
            sum_nodes = sum_nodes+add_nodes

            for transform_wrapper in generator:
                transform = transform_wrapper[0]
                nbnodes = transform.shape[0]
                nbneurons_in = transform.shape[1]
                nbneurons_out = transform.shape[3]

                # transform = transform.reshape(nbnodes * nbneurons_in, nbnodes * nbneurons_out)
                # relevance = relevance.reshape([nbnodes * nbneurons_out, 1])
                relevance = (transform.reshape(nbnodes * nbneurons_in, nbnodes * nbneurons_out) @ relevance.reshape([nbnodes * nbneurons_out, 1])).reshape(nbnodes, nbneurons_in)

            graph_node_weights = relevance.sum(axis=1)
            node_weights_list.append(graph_node_weights)
        node_weights = torch.cat(node_weights_list)

        res_weights = self.node_attn_to_edge_attn(node_weights, data.edge_index) if hasattr(data, 'edge_label') else node_weights
        res_weights = self.min_max_scalar(res_weights)

        return -1, {}, original_clf_logits, res_weights.sigmoid()

    def forward_pass(self, data, epoch, do_sampling, fp_16=True):
        # return self._new_forward_pass(data, epoch, do_sampling, fp_16=True)
        original_clf_logits = self.activations_and_grads(data)
        if data.x.shape[0] > 1400:
            res_weights = torch.ones(data.x.shape[0], )
            print(f"Too many nodes ({data.x.shape[0]}) in this graph, cannot process")
            return -1, {}, original_clf_logits, res_weights.sigmoid()
        # original_clf_logits = self.clf(data)
        # H_end, transforms = self.get_H_transforms(data, gammas)
        H_end = self.activations_and_grads.activations[-1].data
        H_end = H_end.half() if fp_16 else H_end
        sum_nodes = 0
        node_weights_list = []
        for graph in data.to_data_list():
            add_nodes = graph.num_nodes
            relevance = H_end[sum_nodes:sum_nodes+add_nodes, :]
            transforms = self.get_single_graph_transforms(graph, self.gammas, start=sum_nodes, fp_16=fp_16)
            sum_nodes = sum_nodes+add_nodes
            for transform in reversed(transforms):
                # einsum slow
                # relevance_subgraph = torch.einsum('ijkl,kl->ij', transform, mask @ relevance_subgraph)
                nbnodes = transform.shape[0]
                nbneurons_in = transform.shape[1]
                nbneurons_out = transform.shape[3]

                def sparse_reshape(sparse_tensor, size):
                    mask = torch.sparse_coo_tensor(indices=sparse_tensor.indices(), 
                    values=torch.ones(sparse_tensor._nnz(),device=sparse_tensor.device, dtype=torch.bool),
                    size=sparse_tensor.size())

                    mask = mask.to(torch.int64).to_dense().reshape(size)

                    sparse_tensor = torch.sparse_coo_tensor(indices=mask.to_sparse().indices(), values=sparse_tensor.values(),size=size)
                    return sparse_tensor.coalesce()

                transform = transform.reshape(nbnodes * nbneurons_in, nbnodes * nbneurons_out)
                relevance = relevance.reshape([nbnodes * nbneurons_out, 1])
                relevance = (transform @ relevance).reshape(nbnodes, nbneurons_in)
                
                #! This has been for sparse matrix version
                # relevance = relevance.to_sparse_coo()
                # size = (nbnodes * nbneurons_in, nbnodes * nbneurons_out)
                # transform = sparse_reshape(transform, size=size)
                # relevance = sparse_reshape(relevance, size=(nbnodes * nbneurons_out, 1))
                # relevance = sparse_reshape(transform @ relevance, size=(nbnodes, nbneurons_in))

            # relevance = relevance.to_dense()
            graph_node_weights = relevance.sum(axis=1)
            node_weights_list.append(graph_node_weights)
        node_weights = torch.cat(node_weights_list)

        res_weights = self.node_attn_to_edge_attn(node_weights, data.edge_index) if hasattr(data, 'edge_label') else node_weights

        res_weights = self.min_max_scalar(res_weights)

        return -1, {}, original_clf_logits, res_weights.sigmoid()

    @staticmethod
    def get_adj(data):
        adj = torch.eye(data.num_nodes, device=data.edge_index.device)
        for i, j in data.edge_index.T:
            adj[i, j] = 1
            adj[j, i] = 1
        return adj

    def get_last_nth_transform(self, graph, gammas, start, n=-1):
    # 首先，计算生成器的长度
        length = self.get_generator_length()
        # 如果n大于生成器长度，返回None或抛出异常
        if n > length or n <= 0:
            raise ValueError("Wrong index")
        # 计算倒数第n个值的索引（正向计数）
        target_index = length - n
        
        # 重新构造或获取生成器
        generator = self.get_transforms_generator(graph, gammas, start)  # 需要一种方法来重新获得或重置生成器
        # 遍历生成器到达目标索引
        for i, value in enumerate(generator):
            if i == target_index:
                return value[0]

    def get_generator_length(self):
        activations_list = [a.data for a in self.activations_and_grads.activations] 
        return len(activations_list)

    def reverse_transforms_generator(self, graph, gammas, start, fp_16=False):
        transforms = [0]
        activations_list = [a.data for a in self.activations_and_grads.activations]
        weight_list = [g.data for g in self.activations_and_grads.weights]
        bias_list = [b.data for b in self.activations_and_grads.biases]
        A = self.get_adj(graph)
        for W, b, H, gamma in reversed(list(zip(weight_list, bias_list, activations_list, gammas))):
            torch.cuda.empty_cache()
            W = W + gamma * W.clamp(min=0)
            b = b + gamma * b.clamp(min=0) + 1e-6
            b = b.half() if fp_16 else b
            H = H[start:start+graph.num_nodes, :]
            neuron_weight = (A.unsqueeze(-1).unsqueeze(-1) * W.unsqueeze(0).unsqueeze(0)).permute(0, 3, 1, 2)
            neuro_feature = H.unsqueeze(0).unsqueeze(0)
            if fp_16:
                neuro_feature = neuro_feature.half()
                neuron_weight = neuron_weight.half()
            transform = (neuron_weight * neuro_feature)
            transform = transform / (transform.sum(axis=0).sum(axis=0) + b).unsqueeze(0).unsqueeze(0)
            transforms[0] = transform
            # 修改开始
            # transforms.append(transform)  # 删除这行
            yield transforms
            # 每次生成同一个列表 但元素不同 使得使用同一片内存
        torch.cuda.empty_cache()

    def get_single_graph_transforms(self, graph, gammas, start, fp_16=False):
        def memory_size(tensor):
            return tensor.element_size() * tensor.nelement() / 1024**3
        def is_sparse(tensor):
            total_elements = tensor.numel()
            non_zero_elements = torch.count_nonzero(tensor)
            sparsity_ratio = 1 - (non_zero_elements / total_elements)

            print(f"非零元素比例: {non_zero_elements / total_elements:.2f}")
            print(f"稀疏比例: {sparsity_ratio:.2f}")

        activations_list = [a.data for a in self.activations_and_grads.activations]
        weight_list = [g.data for g in self.activations_and_grads.weights]
        bias_list = [b.data for b in self.activations_and_grads.biases]
        transforms = []
        A = self.get_adj(graph)
        for W, b, H, gamma in zip(weight_list, bias_list, activations_list, gammas):
            torch.cuda.empty_cache()
            W = W + gamma * W.clamp(min=0)
            b = b + gamma * b.clamp(min=0) + 1e-6
            b = b.half() if fp_16 else b
            H = H[start:start+graph.num_nodes, :]
            neuron_weight = (A.unsqueeze(-1).unsqueeze(-1) * W.unsqueeze(0).unsqueeze(0)).permute(0, 3, 1, 2)
            neuro_feature = H.unsqueeze(0).unsqueeze(0)

            if fp_16:
                neuro_feature = neuro_feature.half()
                neuron_weight = neuron_weight.half()
            torch.cuda.empty_cache()
            transform = (neuron_weight * neuro_feature) 
            # focus on the computation here especially the 'b'
            transform = transform / (transform.sum(axis=0).sum(axis=0) + b).unsqueeze(0).unsqueeze(0)
            # transform = transform.to_sparse()
            transforms.append(transform)
        torch.cuda.empty_cache()
        return transforms

    def get_H_transforms(self, data, gammas):
        start = 0
        activations_list = [a.cpu().data for a in self.activations_and_grads.activations]
        weight_list = [g.cpu().data for g in self.activations_and_grads.weights]
        bias_list = [b.cpu().data for b in self.activations_and_grads.biases]
        transforms = []

        for W, b, H, gamma in zip(weight_list, bias_list, activations_list, gammas):
            W = W + gamma * W.clamp(min=0)
            b = b + gamma * b.clamp(min=0)
            batch_transforms = []
            start = 0
            for i in range(data.num_graphs):
                graph = data.get_example(i)
                A = self.get_adj(graph)
                neuron_weight = (A.unsqueeze(-1).unsqueeze(-1) * W.unsqueeze(0).unsqueeze(0)).permute(0, 3, 1, 2)
                neuro_feature = H[start:start+graph.num_nodes, :].unsqueeze(0).unsqueeze(0)
                transform = neuron_weight*neuro_feature # 这里没加b？
                transform = transform / (transform.sum(axis=-1).sum(axis=-1)+b).unsqueeze(-1).unsqueeze(-1)
                start = start + graph.num_nodes
                batch_transforms.append(transform)

            batch_transform = self.to_batch_trans(batch_transforms, node_size=data.num_nodes)
            transforms.append(batch_transform)


        return activations_list[-1], transforms

    @staticmethod
    def to_batch_trans(batch_trans, node_size):
        hidden_size = batch_trans[0].shape[1]
        res = torch.zeros([node_size, node_size, hidden_size, hidden_size])
        num_nodes = 0
        for four_dim_tensor in batch_trans:
            four_dim_tensor.permute(0, 2, 1, 3)
            add_nodes = four_dim_tensor.shape[0]
            res[num_nodes:num_nodes+add_nodes, num_nodes:num_nodes+add_nodes, :, :] = four_dim_tensor
            num_nodes = num_nodes + add_nodes
        return res

    @staticmethod
    def node_attn_to_edge_attn(node_attn, edge_index):
        src_attn = node_attn[edge_index[0]]
        dst_attn = node_attn[edge_index[1]]
        edge_attn = (src_attn + dst_attn) / 2
        return edge_attn

class Transforms:
    """ Class for extracting activations and
    registering gradients from targetted intermediate layers """

    def __init__(self, model, target_layers, reshape_transform=None):
        self.model = model
        self.activations = []
        self.weights = []
        self.biases = []
        self.reshape_transform = reshape_transform
        self.handles = []
        for target_layer in target_layers:
            self.handles.append(
                target_layer.register_forward_hook(self.save_activation))
            # Because of https://github.com/pytorch/pytorch/issues/61519,
            # we don't use backward hook to record gradients.
            # self.handles.append(
            #     target_layer.register_forward_hook(self.save_gradient))

    def save_activation(self, module, input, output):
        if isinstance(output, tuple):  # egnn outputs a tuple
            output = output[0]

        reg_fee_dic = RegexMap(module.state_dict(), None)
        activation = output
        weight = reg_fee_dic['weight']
        bias = reg_fee_dic['bias']

        for mod in module.children():
            if isinstance(mod, GINConv):
                weight = torch.ones_like(weight)
        # weight = torch.ones_like(weight)
        if self.reshape_transform is not None:
            activation = self.reshape_transform(activation)
        self.weights.append(weight.detach())
        self.biases.append(bias.detach())
        self.activations.append(activation.detach())

    def save_gradient(self, module, input, output):
        if isinstance(output, tuple):  # egnn outputs a tuple
            output = output[0]

        if not hasattr(output, "requires_grad") or not output.requires_grad:
            # You can only register hooks on tensor requires grad.
            return

        # Gradients are computed in reverse order
        def _store_grad(grad):
            if self.reshape_transform is not None:
                grad = self.reshape_transform(grad)
            self.gradients = [grad.detach()] + self.gradients

        output.register_hook(_store_grad)

    def __call__(self, x):
        self.weights = []
        self.activations = []
        self.biases = []
        return self.model(x)

    def release(self):
        for handle in self.handles:
            handle.remove()

from re import search, I
class RegexMap:
    def __init__(self, n_dic, val):
        self._items = n_dic
        self.__val = val

    def __getitem__(self, key):
        for regex in reversed(self._items.keys()):
            if search(key, regex, I):
                return self._items[regex]
        return self.__val


from .info_utils import *
from .model_utils import *
from .url import *
from .logger import log_epoch, log, update_and_save_best_epoch_res, load_checkpoint, save_checkpoint
# from ..get_data_loaders import get_data_loaders, ScienceDataset


# https://github.com/jacobgil/pytorch-grad-cam/blob/master/pytorch_grad_cam/base_cam.py

import numpy as np
import torch
from torch.autograd import Variable
from torch_geometric.data import Data, Batch
from torch_geometric.nn import MessagePassing
from ..base import BaseRandom


class Grad(BaseRandom):

    def __init__(self, clf, criterion, config) -> None:
        super().__init__()
        self.clf = clf
        self.criterion = criterion
        self.device = next(self.parameters()).device
        # assert signal_class is not None
        self.signal_class = config['signal_class']
        self.activations_and_grads = None

    def start_tracking(self):
        if hasattr(self.clf, 'clf'):
            self.activations_and_grads = ActivationsAndGradients(self.clf.clf, self.target_layers)
        else:
            self.activations_and_grads = ActivationsAndGradients(self.clf, self.target_layers)


    def __del__(self):
        self.activations_and_grads.release()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.activations_and_grads.release()
        if isinstance(exc_value, IndexError):
            # Handle IndexError here...
            print(
                f"An exception occurred in CAM with block: {exc_type}. Message: {exc_value}")
            return True


class GradX(Grad):
    def __init__(self, clf, criterion, config):
        super(GradX, self).__init__(clf, criterion, config)
        if hasattr(clf, 'model'):
            self.target_layers = [clf.model.node_encoder]  # node_embedding
        else:
            self.target_layers = [clf.clf.model.node_encoder]
        # not used if the feature is not categorical
        self.name = 'gradx'

    def forward_pass(self, data, epoch, do_sampling):
        x_level = 'geometric'
        self.clf.eval()
        try:
            is_cat_feat = False
            if x_level == 'graph':
                data.x.requires_grad = True
            else:
                data.pos.requires_grad = True
        except(RuntimeError, ValueError):
            is_cat_feat = True


        original_clf_logits = self.activations_and_grads(data)
        # original_clf_logits = self.clf(data)
        masked_clf_logits = original_clf_logits
        # org_pred = original_clf_logits.sigmoid().int()
        targets = [BinaryClassifierOutputTarget(self.signal_class)] * original_clf_logits.shape[0]
        # targets = [BinaryClassifierOutputTarget(item) for item in org_pred]

        self.clf.zero_grad()
        loss = sum([target(output) for target, output in zip(targets, original_clf_logits)])
        loss.backward(retain_graph=True)
        loss_dict = {'loss': loss.item(), 'pred': loss.item()}

        if is_cat_feat:
            grad = self.activations_and_grads.gradients[0].squeeze().to(loss.device)
            value = self.activations_and_grads.activations[0].squeeze().to(loss.device)
            node_att = (grad * value).sum(dim=1)
        else:
            node_att = data.x.grad.norm(dim=1, p=2) if x_level == 'graph' else data.pos.grad.norm(dim=1, p=2)
                # data.pos.grad.norm(dim=1, p=2)

        res_weights = self.node_attn_to_edge_attn(node_att, data.edge_index) if hasattr(data, 'edge_label') else node_att
        # res_weights = node_att
        res_weights = self.min_max_scalar(res_weights)

        return -1, loss_dict, original_clf_logits, res_weights.reshape(-1)


class GradCAM(Grad):
    def __init__(self, clf, criterion, config):
        super(GradCAM, self).__init__(clf, criterion, config)
        if hasattr(clf, 'model'):
            self.target_layers = [clf.model.convs[-1]]
        else:
            self.target_layers = [clf.clf.model.convs[-1]]
        self.name = 'gradcam'

    def forward_pass(self, data, epoch, do_sampling):
        self.clf.eval()
        original_clf_logits = self.activations_and_grads(data)
        masked_clf_logits = original_clf_logits
        targets = [BinaryClassifierOutputTarget(self.signal_class)] * original_clf_logits.shape[0]

        self.clf.zero_grad()
        loss = sum([target(output) for target, output in zip(targets, original_clf_logits)])
        loss.backward(retain_graph=True)
        loss_dict = {'loss': loss.item(), 'pred': loss.item()}


        cam_per_layer = self.compute_cam_per_layer()
        node_att = torch.tensor(self.aggregate_multi_layers(cam_per_layer), device=loss.device)

        # res_weights = self.node_attn_to_edge_attn(node_att, data.edge_index) if x_level == "graph" else node_att
        res_weights = self.node_attn_to_edge_attn(node_att, data.edge_index) if hasattr(data,
                                                                                    'edge_label') else node_att
        res_weights = self.min_max_scalar(res_weights)

        return -1, loss_dict, original_clf_logits, res_weights.reshape(-1)

    def get_cam(self, activations: torch.Tensor, grads: torch.Tensor) -> np.ndarray:
        assert activations.min() >= 0.0  # rectified
        cam = (grads * activations).sum(1)
        return cam

    def aggregate_multi_layers(self, cam_per_target_layer: np.ndarray) -> np.ndarray:
        cam_per_target_layer = np.stack(cam_per_target_layer, axis=1)
        result = np.mean(cam_per_target_layer, axis=1)
        return result

    def compute_cam_per_layer(self) -> np.ndarray:
        activations_list = [a.cpu().data.numpy() for a in self.activations_and_grads.activations]
        grads_list = [g.cpu().data.numpy() for g in self.activations_and_grads.gradients]

        cam_per_target_layer = []
        # Loop over the saliency image from every layer
        for i in range(len(self.target_layers)):
            layer_activations = None
            layer_grads = None
            if i < len(activations_list):
                layer_activations = activations_list[i]
            if i < len(grads_list):
                layer_grads = grads_list[i]

            cam = self.get_cam(layer_activations, layer_grads)
            cam_per_target_layer.append(cam)
        return cam_per_target_layer


class InterGrad(Grad):

    def __init__(self, clf, criterion, config):
        super(InterGrad, self).__init__(clf, criterion, config)
        if hasattr(clf, 'model'):
            self.target_layers = [clf.model.node_encoder]
        else:
            self.target_layers = [clf.clf.model.node_encoder]
        self.name = 'inter_grad'

    def forward_pass(self, data, baseline=0, steps=20, **kwargs):
        # node_base = torch.zeros_like(data.x) if node_base == None else node_base
        # edge_base = torch.zeros_like(data.edge_attr) if edge_base == None else edge_base
        x_level = 'geometric'
        self.clf.eval()
        original_clf_logits = self.activations_and_grads(data)
        scales = [baseline + (float(i) / steps) * (1 - baseline) for i in range(1, steps + 1)]
        node_grads = []
        step_len = float(1 - baseline) / steps
        for scale in reversed(scales):
            if x_level == 'geometric':
                new_data_list = []
                for graph in data.to_data_list():
                    new_instance = graph.clone()
                    new_instance.pos = scale * new_instance.pos
                    new_data_list += [new_instance]
            else:
                new_data_list = []
                for graph in data.to_data_list():
                    new_instance = graph.clone()
                    new_instance.x = scale * new_instance.x
                    new_data_list += [new_instance]
                # new_data_list = [Data(x=scale * graph.x, edge_index=graph.edge_index, y=graph.y) for graph in data.to_data_list()]
            follow_batch = None if not hasattr(data, 'x_lig') else ['x_lig']
            new_data = Batch.from_data_list(new_data_list, follow_batch=follow_batch)
            if x_level == 'geometric':
                new_data.pos.requires_grad = True
            else:
                try:
                    is_cat_feat = False
                    new_data.x.requires_grad = True
                except(RuntimeError, ValueError):
                    is_cat_feat = True
            pred = self.activations_and_grads(new_data)
            # pred = self.clf(new_data, edge_attr=data.edge_attr)
            sum(pred).backward(retain_graph=True)
            if x_level == 'geometric':
                score = new_data.pos.grad.norm(dim=1, p=2)
            else:
                if is_cat_feat:
                    grad = self.activations_and_grads.gradients[0].squeeze().to(pred.device)
                    # value = self.activations_and_grads.activations[0].squeeze().to(loss.device)
                    score = grad.norm(dim=1)
                else:
                    score = new_data.x.grad.norm(dim=1, p=2)
            # score = pow(new_data.pos.grad, 2).sum(dim=1).cpu().numpy()
            node_grads.append((score * step_len).cpu())

            self.clf.zero_grad()
        node_grads = torch.tensor(np.stack(node_grads), device=data.x.device).sum(axis=0)
        # node_grads[node_grads < 0] = 1e-16
        # node_imp = (node_grads - min(node_grads)) / (max(node_grads) - min(node_grads))

        res_weights = self.node_attn_to_edge_attn(node_grads, data.edge_index) if hasattr(data, 'edge_label') else node_grads
        res_weights = self.min_max_scalar(res_weights)

        return -1, {}, original_clf_logits, res_weights.reshape(-1)


class BinaryClassifierOutputTarget:
    def __init__(self, category):
        self.category = category

    def __call__(self, model_output):
        if self.category == 1:
            sign = 1
        else:
            sign = -1
        return model_output * sign


class ActivationsAndGradients:
    """ Class for extracting activations and
    registering gradients from targetted intermediate layers """

    def __init__(self, model, target_layers, reshape_transform=None):
        self.model = model
        self.gradients = []
        self.activations = []
        # self.emb_x = torch.tensor(0)
        self.reshape_transform = reshape_transform
        self.handles = []
        # self.handles.append(model.backbone.node_encoder.register_forward_hook(self.save_activation))
        # self.handles.append(model.backbone.node_encoder.register_forward_hook(self.save_gradient))
        for target_layer in target_layers:
            self.handles.append(
                target_layer.register_forward_hook(self.save_activation))
            # Because of https://github.com/pytorch/pytorch/issues/61519,
            # we don't use backward hook to record gradients.
            self.handles.append(
                target_layer.register_forward_hook(self.save_gradient))

    def save_activation(self, module, input, output):
        if isinstance(output, tuple):  # egnn outputs a tuple
            output = output[0]
        # if isinstance(module, nn.Embedding):
        #     self.emb_x = output
        #     return

        activation = output
        if self.reshape_transform is not None:
            activation = self.reshape_transform(activation)
        self.activations.append(activation.cpu().detach())

    def save_gradient(self, module, input, output):
        if isinstance(output, tuple):  # egnn outputs a tuple
            output = output[0]
        # if isinstance(module, nn.Embedding):
        #     def _store_embx_grad(grad):
        #         self.emb_x.grad = grad
        #     output.register_hook(_store_embx_grad)
        #     return

        if not hasattr(output, "requires_grad") or not output.requires_grad:
            # You can only register hooks on tensor requires grad.
            return

        # Gradients are computed in reverse order
        def _store_grad(grad):
            if self.reshape_transform is not None:
                grad = self.reshape_transform(grad)
            self.gradients = [grad.cpu().detach()] + self.gradients

        output.register_hook(_store_grad)

    def __call__(self, x):
        self.gradients = []
        self.activations = []
        return self.model(x)

    def release(self):
        if self.handles:
            for handle in self.handles:
                handle.remove()



import yaml
import json
import re
import argparse
from tqdm import tqdm
from pathlib import Path
import torch
import torch.nn as nn
from torch.nn import functional as F
from eval import FidelEvaluation, AUCEvaluation, PrecEvaluation
from get_model import Model
from baselines import *
from utils import to_cpu, log_epoch, get_data_loaders, set_seed, load_checkpoint, ExtractorMLP, get_optimizer, sum_fid_score
from utils import inherent_models, post_hoc_explainers, post_hoc_attribution, name_mapping
import torchmetrics
import numpy as np
import pandas as pd
import warnings
from datetime import datetime
warnings.filterwarnings("ignore")


def eval_one_batch(baseline, data, epoch, return_attn=True):
    # with torch.set_grad_enabled(baseline.name in ['gradcam', 'gradx', 'inter_grad', 'gnnexplainer']):
    with torch.set_grad_enabled(baseline.name in ['gradx', 'gradcam', 'inter_grad', 'gnnexplainer']):
        baseline.eval()
        # baseline.extractor.eval() if hasattr(baseline, 'extractor') else None
        # baseline.clf.eval()

        # do_sampling = True if phase == 'valid' and baseline.name == 'pgexplainer' else False # we find this is better for BernMaskP
        if return_attn:
            loss, loss_dict, infer_clf_logits, node_attn = baseline.forward_pass(data, epoch=epoch, do_sampling=False)
        else:
            infer_clf_logits = baseline.forward(data) if baseline.name in inherent_models else baseline.clf(data)
            loss_dict, node_attn = {}, None

        return loss_dict, to_cpu(infer_clf_logits), to_cpu(node_attn)



def eval_one_epoch(baseline, data_loader, epoch, phase, seed, signal_class, metric_list=None, attn_dir=None):
    return_attn = False if attn_dir else True
    pbar = tqdm(data_loader)
    [eval_metric.reset() for eval_metric in metric_list] if metric_list else None

    # treshhold_list = np.linspace(0.1, 1, 10)
    # for num in range(1, 10):
    #     exec(f"clf_acc_{num} = torchmetrics.Accuracy(task='binary', threshold={0.9+num/100})")

    clf_ACC, clf_AUC = torchmetrics.Accuracy(task='binary', threshold=0.5), torchmetrics.AUROC(task='binary')
    for idx, data in enumerate(pbar):
        torch.cuda.empty_cache()
        # data = negative_augmentation(data, data_config, phase, data_loader, idx, loader_len)
        loss_dict, clf_logits, attn = eval_one_batch(baseline, data.to(baseline.device), epoch, return_attn=return_attn)

        if return_attn is False:
            # if the seed are bigger or equal than 99, the attn_dir will be a list of file names.
            if seed > 99:
                b_seed = seed - 100
                seed_attn_dir = attn_dir[b_seed]
                all_df = pd.read_csv(seed_attn_dir)
                ens_cols = [col for col in all_df.columns if col.isdigit()]
                # ens_cols = [str(i) for i in range(10)]
                ensemble_df = all_df[ens_cols].mean(axis=1)
                attn = torch.tensor(ensemble_df[all_df['batch_idx'] == idx].to_numpy())
            elif seed == 99:
                multi_attn = None
                weight_list = []
                #// pattern = r'\/([^\/]+)\/bs\d+_ms\[\d+\]_attns\.csv'
                #// method = re.findall(pattern, str(attn_dir[0]))[0]
                _, setting, method, *_ = str(attn_dir[0]).split('/') 

                #~ TODO: add the logic of weighted
                # col can be "exp_auc", "ACCfidpos"
                def add_weight_list(cur_seed: int, setting, col='ACCfidpos'):
                    #* find the target df
                    regex = r"^attribution.*erm$"
                    root_path = Path("result", setting, "post-inherent/ensenmle_exp_auc_prec@20_fid+pos_fid-pos/")
                    target_dir_list = [dir for dir in root_path.iterdir() if dir.is_dir() and re.search(regex, dir.name)]
                    # print(target_dir_list)
                    target_dir = target_dir_list[0]
                    target_file = [file for file in target_dir.iterdir() if file.is_file() and f"bseed{cur_seed}" in file.name][0]
                    temp = pd.read_csv(target_file, index_col=0)

                    #! Important selection
                    result_df = temp.loc[temp['exp_method']==method]
                    if len(result_df) == 1:
                        weight_list.append(float(result_df[col]))
                    elif len(result_df) > 10:
                        row = result_df.index.str.contains(f'erm_{cur_seed}')
                        weight_list.append(float(result_df[row][col]))
                    else:
                        print(target_file)
                        print(temp)
                        print(method)
                        raise NotImplementedError(f"Unknown df length in 1-10, df: {result_df}")
                    return
                #* The number of digits in the name that indicates the number of explainer seed. Only used for subgraphx.
                pre_num = len(re.findall(r'\d+', str(attn_dir[0]))) - 1 if method == 'subgraphx' else 10
                for i in range(pre_num):
                    seed_attn_dir = attn_dir[i]
                    # seed_attn_dir = str(attn_dir).replace(f'bs{b_seed}', f'bs{i}')
                    seed_df = pd.read_csv(seed_attn_dir)
                    name = 'attn' if 'attn' in seed_df else str(i)
                    #! Error: Can not convert numpy.object_
                    seed_attn = torch.tensor(seed_df[name][seed_df['batch_idx'] == idx].to_numpy()).unsqueeze(1)
                    multi_attn = seed_attn if multi_attn is None else torch.cat([multi_attn, seed_attn], dim=1)
                    
                    #! Remember to uncomment it
                    add_weight_list(i, setting=setting)

                #! This is for testing the new Agg Strategy
                weights = torch.tensor(weight_list, dtype=torch.float64).view(-1, 1)
                if (weights > 0).any():
                    weights = torch.clamp(weights, min=0)
                else:
                    weights = torch.nn.functional.softmax(weights, dim=0)
                # weights = (weights + 1) / 2 
                # normalized_w = torch.nn.functional.softmax(weights, dim=0)
                if True:
                    attn = torch.matmul(multi_attn, weights) / weights.sum()                
                else:
                    attn = torch.mean(multi_attn, dim=1) ## shape == (N, 10) 

            else:
                seed_df = pd.read_csv(attn_dir)
                name = 'attn' if 'attn' in seed_df else str(seed)
                attn = torch.tensor(seed_df[name][seed_df['batch_idx'] == idx].to_numpy())


        ex_labels, clf_labels, data = to_cpu(data.node_label), to_cpu(data.y), data.cpu()
        eval_dict = {metric.name: metric.collect_batch(ex_labels, attn, data, signal_class, 'geometric')
                     for metric in metric_list}
        eval_dict.update({'clf_acc': clf_ACC(clf_logits, clf_labels), 'clf_auc': clf_AUC(clf_logits, clf_labels)})

        # acc_dict = {}
        # for num in range(1, 10):
        #     exec(f"acc_dict['clf_acc_{num}'] = clf_acc_{num}(clf_logits, clf_labels)")
        # print('='*80, acc_dict, '='*80)

        desc = log_epoch(seed, epoch, phase, loss_dict, eval_dict, phase_info='dataset')
        pbar.set_description(desc)
        # compute the avg_loss for the epoch desc

    epoch_dict = {eval_metric.name: eval_metric.eval_epoch() for eval_metric in metric_list} if metric_list else {}
    epoch_dict.update({'clf_acc': clf_ACC.compute().item(), 'clf_auc': clf_AUC.compute().item()})
    sum_fid_score(epoch_dict)
    # log_epoch(seed, epoch, phase, avg_loss_dict, epoch_dict, writer)

    return epoch_dict


def test(config, method_name, exp_method, model_name, backbone_seed, method_seed, loaders, dataset, config_dir, device, metric_str, csv_dir):
    set_seed(backbone_seed, method_name=method_name, backbone_name=model_name)
    writer = None
    model_dir = config_dir / method_name
    log_dir = config_dir / "post-inherent"
    log_dir.mkdir(parents=True, exist_ok=True)
    print('The logging directory is', log_dir), print('=' * 80)

    batch_size = config['optimizer']['batch_size']
    if exp_method == 'gnnlrp' and dataset == 'plbind':
        batch_size = 1
    data_config = config['data']
    loaders, test_set, dataset = get_data_loaders(dataset_name, batch_size, data_config, dataset_seed=0, num_workers=0, device=device)
    signal_class = dataset.signal_class

    clf = Model(model_name, config[model_name],method_name, config[method_name], dataset).to(device)
    backbone_config = config[model_name]
    method_config = config[method_name]

    extractor = ASAPooling(backbone_config['hidden_size'], ratio=method_config['casual_ratio'], dropout=method_config['dropout_p']) \
        if method_name == 'asap' else ExtractorMLP(backbone_config['hidden_size'], method_config, config['data'].get('use_lig_info', False)) \
        if method_name in inherent_models else nn.Identity()
    extractor = extractor.to(device)
    criterion = F.binary_cross_entropy_with_logits
    map_location = torch.device('cpu') if not torch.cuda.is_available() else device
    # establish the clf model and the metrics
    backbone = eval(name_mapping[method_name])(clf, extractor, criterion, config[method_name]) \
        if method_name in inherent_models else clf

    # initialize extractor of pgexplainer
    if exp_method == 'pgexplainer':
        pg_extractor = ExtractorMLP(backbone_config['hidden_size'], config['pgexplainer'], config['data'].get('use_lig_info', False))
        pg_extractor = pg_extractor.to(device)
        model = PGExplainer(backbone, pg_extractor, criterion, config['pgexplainer'])
        if clf_method == 'erm':
            load_checkpoint(model, config_dir / 'pgexplainer', model_name=exp_method, backbone_seed=backbone_seed, seed=method_seed,
                        map_location=map_location) if method_seed < 10 else print(f'NO PGExplainer loaded for ensemble seed {method_seed}.')
        else:
            print(f"Prepare to train PGExplainer for Inherent Model {clf_method}.")

            pass

    # initialize erm or inherent_models from ckpt
    assert load_checkpoint(backbone, model_dir, model_name=method_name, seed=backbone_seed, map_location=map_location) \
        # if exp_method != 'pgexplainer' else True

    if 'label_perturb' in exp_method:
        model = LabelPerturb(backbone, mode=int(exp_method[-1]))
    elif 'self' == exp_method:
        model = backbone
    elif exp_method == 'pgexplainer':
        if clf_method == 'erm':
            pass  # solved in the former part to avoid overwrite the erm although they should be the same
        else:
            print("Train PGExplainer...")

    else:
        model = eval(name_mapping[exp_method])(backbone, criterion, config[exp_method])

    # metric_list = [AUCEvaluation()] + [FidelEvaluation(backbone, i/10) for i in range(2, 9)] + \
    #  [FidelEvaluation(backbone, i/10, instance='pos') for i in range(2, 9)] + \
    metric_list = []
    for one_metric in metric_str:
        if 'exp_auc' == one_metric:
            metric_list += [AUCEvaluation()]
        elif re.compile(r'prec').search(one_metric):
            _, k = one_metric.split("@")
            metric_list += [PrecEvaluation(int(k))]
        elif 'fid+pos' == one_metric:
            metric_list += [FidelEvaluation(backbone, i / 10, instance='pos', type='acc', symbol='+') for i in reversed(range(2, 9))]
        elif 'fid-pos' == one_metric:
            metric_list += [FidelEvaluation(backbone, i / 10, instance='pos', type='acc', symbol='-') for i in reversed(range(2, 9))]
        else:
            print('='*80, f'The unknown metric is set as {one_metric}', '='*80)
            # raise ValueError(f"Metric name {one_metric} besides exp_auc/pos_fid/neg_fid is not expected")

    model.start_tracking() if 'grad' in exp_method or exp_method == 'gnnlrp' else None
    set_seed(method_seed, method_name=method_name, backbone_name=model_name)
    # metric_names = [i.name for i in metric_list] + ['clf_acc', 'clf_auc']
    # train_dict = eval_one_epoch(model, loaders['train'], 1, 'train', backbone_seed, signal_class, writer, metric_list)
    # valid_dict = eval_one_epoch(model, loaders['valid'], 1, 'valid', backbone_seed, signal_class, writer, metric_list)
    # if exp_method == 'pgexplainer' and method_name != 'erm':
    #     optimizer = get_optimizer(clf, extractor, config['optimizer'], method_name, warmup=False)
    #     raise NotImplementedError('PGExplainer for inherent models are not implemented yet.')
    test_dict = eval_one_epoch(model, loaders['test'], 1, 'test', method_seed, signal_class, metric_list, attn_dir=csv_dir)

    if method_seed == 99:
        exp_method += '_mul_ens'
        method_name += '_mul_ens'
    elif method_seed >= 100:
        exp_method += '_one_ens'
        method_name += '_one_ens'
    test_dict.update({'exp_method': exp_method, 'clf_method': method_name})
    return test_dict


def save_multi_methods(multi_methods_res, indexes, seed_info, metric_list, exp_list, mod_list, only_ens=False):
    # save_seeds = seeds +
    # indexes = ['_'.join([method_name, str(seed), item]) for seed in seeds for item in ['avg', 'std']]
    # from itertools import product
    # indexes = ['_'.join(item) for item in product(methods, seeds)]
    df = pd.DataFrame(multi_methods_res, index=indexes)

    #! Remember to delete the "_only_ens"
    suffix = "_only_ens" if only_ens else ""
    day_dir = Path('result') / config_name / 'post-inherent' / '_'.join(metric_list) / ('_'.join(exp_list+mod_list) + suffix)
    day_dir.mkdir(parents=True, exist_ok=True)
    csv_dir = day_dir / ('_'.join([f'bseed{seed_info}']+exp_list+mod_list) + '.csv')
    with open(csv_dir, mode='w') as f:
        df.to_csv(f, lineterminator="\n", header=f.tell() == 0)
    return df

def get_avg_std_report(reports, around=2):
    # print(reports)
    one_report = reports[0]
    str_dicts = {k: one_report[k] for k in one_report if isinstance(one_report[k], str)}
    all_keys = {k: [] for k in one_report if not isinstance(one_report[k], str)}
    for report in reports:
        for k in all_keys:
            all_keys[k].append(report[k])
    avg_report, std_report, avg_std_report = str_dicts.copy(), str_dicts.copy(), str_dicts.copy()
    avg_report.update({k: np.mean(v) for k, v in all_keys.items()})
    std_report.update({k: np.std(v) for k, v in all_keys.items()})
    avg_std_report.update({k: fr'{(100*np.mean(v)):.{around}f} \pm {(100*np.std(v)):.{around}f}' for k, v in all_keys.items()})
    return avg_report, std_report, avg_std_report

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='EXP')
    parser.add_argument('-d', '--dataset', type=str, help='dataset used', default='actstrack')
    parser.add_argument('--clf_method', type=str, nargs="+", help='classification method used', default=['erm'])
    parser.add_argument('--exp_method', type=str, nargs="+", help='explanation method used', default=['self'])
    parser.add_argument('-b', '--backbone', type=str, help='backbone used', default='egnn')
    parser.add_argument('--cuda', type=int, help='cuda device id, -1 for cpu', default=-1)
    parser.add_argument('--note', type=str, help='note in log name', default='')
    parser.add_argument('--bseed', type=int, help='random seed for training backbone', default=0)
    # parser.add_argument('--mseeds', type=int, nargs="+", help='random seed for explainer', default=list(range(10)))
    parser.add_argument('--metrics', type=str, nargs="+", default=['exp_auc', 'fid+pos', 'fid-pos'])
    parser.add_argument('--ensemble', type=str, default='exp_auc fidelity', choices=['exp_auc', 'fidelity', 'exp_auc fidelity', ''])
    parser.add_argument('--use_csv', action="store_true")
    parser.add_argument('--gpu_ratio', type=float, help='gpu memory ratio', default=None)


    args = parser.parse_args()
    # sub_metric = 'avg_loss'
    print(args)
    if args.gpu_ratio is not None:
        torch.cuda.set_per_process_memory_fraction(args.gpu_ratio, args.cuda)

    dataset_name, model_name, cuda_id, note = args.dataset, args.backbone, args.cuda, args.note

    device = torch.device(f'cuda:{cuda_id}' if cuda_id >= 0 and torch.cuda.is_available() else 'cpu')

    config_name = '_'.join([model_name, dataset_name])
    # sub_dataset_name = '_' + dataset_name.split('_')[1] if len(dataset_name.split('_')) > 1 else ''
    config_path = Path('./configs') / f'{config_name}.yml'
    config = yaml.safe_load((config_path).open('r'))

    # the directory is like egnn_actstrack / lri_bern
    config_dir = Path('log') / config_name
    multi_method_res, indexes, exp_methods, clf_methods = [], [], ["label_perturb1", "label_perturb0", "self"], []
    #! remember to commet this
    if True:
        exp_methods = []
    #'self', 'label_perturb1', 'label_perturb0',

    backbone_seed = args.bseed

    #! this is explained in paper 
    if dataset_name == "plbind":
        args.metrics.remove('exp_auc')
        args.metrics.insert(0, 'prec@20')
        # print("!!has changed the metric!!")
        # post_hoc_attribution = ['gradcam', 'gradx', 'inter_grad', ]
    for item in args.exp_method:
        exp_methods += post_hoc_attribution if item == 'attribution' else post_hoc_explainers if item == 'explainer'\
                else [item]
    for item in args.clf_method:
        clf_methods += inherent_models if item == 'inherent' else [item] if item in inherent_models else ['erm']

    # if trigger ensemble
    if args.ensemble == 'fidelity':
        assert set(post_hoc_explainers) >= set(exp_methods)

    #! Totally different logic: just attn ensemble
    print('=' * 80)
    print(f"The clf_method: {clf_methods}.")
    print(f"The exp_method: {exp_methods}.")
    print(f"Ensemble: {args.ensemble}.")

    dataset, loaders = args.dataset, {'test': 0}

    for exp_method, clf_method in [(i, j) for i in exp_methods for j in clf_methods]:
        torch.cuda.empty_cache()
        if exp_method == 'self' and clf_method == 'erm':
            continue
        print('=' * 80)
        print(f'Using {exp_method} to explain {clf_method}.')
        mseeds_res = []
        # cuda_id = backbone_seed % 5
        if exp_method in (post_hoc_attribution + ['self']):
            mseeds = [0]
        else: # (exp_method in posthoc_explainer)
            if clf_method in inherent_models:
                mseeds = [backbone_seed]
            else: # clf_method == 'erm'
               mseeds = list(range(10)) 

        # the results of inherent explained by posthoc+self do not use the csv file with stored attn
        # the csv_dir is only used for the post-hoc explaining erm
        use_csv = args.use_csv and (clf_method == 'erm') and ('label_perturb' not in exp_method)
        # out of efficiency, we only compute 3/5 seeds for subgraphx in dataset ActsTrack/SynMol
        if exp_method == 'subgraphx':
            mseeds = list(range(3)) if dataset_name == 'actstrack' else list(range(5))

        #! remember to comment/uncomment
        # for method_seed in mseeds:
        #     csv_dir = config_dir / exp_method / f"bs{backbone_seed}_ms{mseeds}_attns.csv" if use_csv else None
        #     test_report = test(config, clf_method, exp_method, model_name, backbone_seed, method_seed, loaders, dataset, config_dir, device, metric_str=args.metrics, csv_dir=csv_dir)

        #     indexes += [f'{exp_method}_{clf_method}_{method_seed}']
        #     mseeds_res += [test_report] if method_seed < 10 else []

        # avg_report, std_report, avg_std_report = get_avg_std_report(mseeds_res)

        # mseeds_res += [avg_report, avg_std_report] if len(mseeds) > 1 else []
        # indexes += [f'{exp_method}_{clf_method}_avg', f'{exp_method}_{clf_method}_std'] if len(mseeds) > 1 else []
        # multi_method_res += mseeds_res

        # print(json.dumps(avg_std_report, indent=4))
    #! Remember to uncomment them and set only_ens
    only_ens = True

    # for ensemble to add some big seed as indicator
    # 99 indicate the multi model ensemble; 100+ indicates the one model ensemble and the last digit is the backbone seed
    if_ensemble = 'ensenmle' if args.ensemble else ''
    print('=' * 80, '\nBeginning Ensemble!\n', '=' * 80) if if_ensemble else print('Done!')
    for exp_method, clf_method in [(i, j) for i in exp_methods for j in clf_methods]:
        ens_seeds = []
        if exp_method == 'self' and clf_method in inherent_models:
            ens_seeds += [99] if 'exp_auc' in args.ensemble else []
            csv_dirs = [config_dir / clf_method / f"bs{bs}_ms{[bs]}_attns.csv" for bs in range(10)]
        elif exp_method in post_hoc_attribution and clf_method == 'erm':
            ens_seeds += [99] if 'exp_auc' in args.ensemble else []
            csv_dirs = [config_dir / exp_method / f"bs{bs}_ms[0]_attns.csv" for bs in range(10)]
        elif exp_method == 'subgraphx' and clf_method == 'erm':
            ens_seeds += [99] if 'exp_auc' in args.ensemble else []
            ens_seeds += [100 + backbone_seed] if 'fidelity' in args.ensemble else []
            seed_list = list(range(5)) if dataset_name == 'synmol' else list(range(3))
            csv_dirs = [config_dir / exp_method / f"bs{bs}_ms{seed_list}_attns.csv" for bs in range(10)]
        elif exp_method in post_hoc_explainers and clf_method == 'erm':
            ens_seeds += [99] if 'exp_auc' in args.ensemble else []
            ens_seeds += [100+backbone_seed] if 'fidelity' in args.ensemble else []
            csv_dirs = [config_dir / exp_method / f"bs{bs}_ms{list(range(10))}_attns.csv" for bs in range(10)]
        else:
            print(f"No Processing for Explanation Method {exp_method} and Classification Method {clf_method}.")
            continue
        for ens_seed in ens_seeds:
            test_report = test(config, clf_method, exp_method, model_name, backbone_seed, ens_seed, loaders, dataset,
                               config_dir, device, metric_str=args.metrics, csv_dir=csv_dirs)
            multi_method_res += [test_report]
            post_fix = 'multi_ens' if ens_seed == 99 else 'one_ens'
            indexes += [f'{exp_method}_{clf_method}_{post_fix}']

    result = save_multi_methods(multi_method_res, indexes, args.bseed, [if_ensemble]+args.metrics, args.exp_method, args.clf_method, only_ens=only_ens)
    print(result)


import torch
import numpy as np
from datetime import datetime
from sklearn.metrics import roc_auc_score
import torch.nn.functional as F
import os
from pathlib import Path
from functools import partial
from .info_utils import inherent_models, post_hoc_explainers, post_hoc_attribution
def to_item(tensor):
    if tensor is None:
        return None
    elif isinstance(tensor, torch.Tensor):
        return tensor.item()
    else:
        return tensor


def log_epoch(epoch, phase, loss_dict, log_dict=None, seed=None, writer=None, phase_info='phase'):
    comb_dict = dict(loss_dict, **log_dict) if log_dict is not None else loss_dict
    # print(comb_dict) if log_dict else None
    des_phase = phase + ' ' if phase in ['test', 'warm'] else phase  # align tqdm desc bar
    if phase_info == 'phase':
        if seed:
            init_desc = f'[Seed {seed}, {des_phase.capitalize()} Epoch {epoch}] '
        else:
            init_desc = f'[{des_phase.capitalize()} Epoch {epoch}] '
    else:
        assert phase_info == 'dataset'  # this api is used in test_sensitivity.py
        init_desc = f'[Seed {seed}, {des_phase.capitalize()} Dataset] '
    info_desc = ' '.join([k + f': {v:.3f}' for k, v in comb_dict.items()])
    # eval_desc, org_clf_acc, org_clf_auc, exp_auc = get_eval_score(epoch, phase, log_dict, writer, batch)
    if writer is not None:
        for metric, value in comb_dict.items():
            writer.add_scalar(metric, value, epoch)
    return init_desc + info_desc


def log(*args):
    print(f'[{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}]', *args)


def update_and_save_best_epoch_res(baseline, best, valid, test, epoch, model_dir, backbone_seed, seed, writer, method_name, main_metric, sub_metric=None):
    valid_metric = f'valid_{main_metric}'
    valid_sub_metric = f'valid_{sub_metric}'
    better_val = valid[main_metric] > best[valid_metric]
    if sub_metric is not None:
        same_val_but_better = (valid[main_metric] == best[valid_metric]) and (valid[sub_metric] < best[valid_sub_metric])

    # update the best_metric
    if better_val:
        init_dict = {'best_epoch': epoch}
        valid_res = {'valid_' + k: v for k, v in valid.items()}
        test_res = {'test_' + k: v for k, v in test.items()}
        best.update({k: v for d in [init_dict, valid_res, test_res] for k, v in d.items()})
        # if model_dir is not None:
        if method_name in inherent_models + ['pgexplainer']:
            save_checkpoint(baseline, model_dir, model_name=method_name, backbone_seed=backbone_seed, seed=seed)
        ## for post-hoc explaining inherent models
        # elif method_name.startswith('pgexplainer_'):
        #     assert backbone_seed == seed
        #     save_checkpoint(baseline, model_dir, model_name=method_name, backbone_seed=backbone_seed, seed=seed)

        if writer is not None:
            for metric, value in best.items():
                writer.add_scalar(f'best_{metric}', value, epoch)

    # print the best result every 10 epochs
    if epoch % 10 == 0:
        more_readable = partial(map, lambda x: x.upper() if x in ['exp', 'clf'] else x)
        init_desc = f"[Seed {seed}, Best Epoch: {best['best_epoch']}] "
        desc = ', '.join(['_'.join(more_readable(k.split('_'))) + f': {v:.3f}' for k, v in best.items() if k != 'best_epoch'])
        print('-' * 80), print(init_desc + desc), print('-' * 80)

    return best, better_val


def load_checkpoint(model, model_dir, model_name, seed, map_location=None, backbone_seed=None, verbose=True):
    load_dir = model_dir / (model_name + str(seed) + '.pt') if model_name in inherent_models + ['erm'] else model_dir / (str(backbone_seed) + model_name + str(seed) + '.pt')
    if not os.path.exists(load_dir):
        print('There is no checkpoint in', load_dir)
    else:
        checkpoint = torch.load(load_dir, map_location=map_location)
        model.load_state_dict(checkpoint['model_state_dict'])

        print('Successfully load checkpoint in', load_dir) if verbose else None
        return True


def save_checkpoint(model, model_dir, model_name, backbone_seed, seed, save_info=None):
    # assert info is not None
    if model_name == 'erm':
        assert backbone_seed == seed
        if save_info is None:
            save_dir = model_dir / (model_name + str(seed) + '.pt')
            torch.save({'model_state_dict': model.state_dict()}, save_dir)
        else:
            epo, acc, auc = save_info['epoch'], save_info['acc'] * 1000, save_info['auc'] * 1000
            (model_dir / 'models').mkdir(parents=True, exist_ok=True)
            mid_dir = model_dir / 'models' / f'seed{backbone_seed}-{epo}epo-{int(auc)}auc.pt'
            # final_dir = model_dir / (model_name + str(seed) + '.pt')
            torch.save({'model_state_dict': model.state_dict()}, mid_dir)
            # torch.save({'model_state_dict': model.state_dict()}, final_dir)

    elif model_name in inherent_models:
        if save_info is None:
            assert backbone_seed == seed
            save_dir = model_dir / (model_name + str(seed) + '.pt')
            torch.save({'model_state_dict': model.state_dict()}, save_dir)
        else:
            epo, acc, auc, xauc = save_info['epoch'], save_info['acc'] * 1000, save_info['auc'] * 1000, save_info['xauc'] * 1000
            #! This use the local path
            _, setting, method_name = str(model_dir).split('/')
            temp_dir = Path('/usr/scratch/jzhu617/suppl_inherent_models/', setting, method_name)
            temp_dir.mkdir(parents=True, exist_ok=True)
            mid_dir = temp_dir / f'seed{backbone_seed}-{epo}epo-{int(auc)}auc-{int(xauc)}xauc.pt'
            # final_dir = model_dir / (model_name + str(seed) + '.pt')
            torch.save({'model_state_dict': model.state_dict()}, mid_dir)
            # torch.save({'model_state_dict': model.state_dict()}, final_dir)

    elif model_name in post_hoc_explainers:
        save_dir = model_dir / (str(backbone_seed) + model_name + str(seed) + '.pt')
        torch.save({'model_state_dict': model.state_dict()}, save_dir)
    # else: # for post-hoc explaining inherent models
    #     save_dir = model_dir / (str(backbone_seed) + model_name + str(seed) + '.pt')
    #     torch.save({'model_state_dict': model.state_dict()}, save_dir)





from .pgexplainer import PGExplainer
from .gnnexplainer import GNNExplainer
from .gnnlrp import GNNLRP
from .orphicx import Orphicx
from .grads import GradX, GradCAM, InterGrad
from .pgmexplainer import PGMExplainer
from .subgraphx import SubgraphX



# 纯文本邮件
import smtplib
from email.mime.text import MIMEText
from email.utils import formataddr
import sys

from_addr = '1831948551@qq.com'
to_addr = '1831948551@qq.com'
password = 'zzeexlwsatdkchdb' # from_addr 需要授权码
# 构造邮件
def send(content):
    msg = MIMEText(content, 'plain', 'utf-8')
    msg['From'] = formataddr(('Host', from_addr)) # 随便打
    msg['To'] = formataddr(('ZHUConv', to_addr)) #随便打
    if 'error' in content:
        msg['Subject'] = 'Program Error' #邮件标题
    else:
        msg['Subject'] = 'Program Done'
    # smtp_server = 'smtp.qq.com'
    # server = smtplib.SMTP_SSL(smtp_server)
    server = smtplib.SMTP_SSL("smtp.qq.com", 465)  # QQ邮箱的SMTP服务需SSL加密，端口为465
    # 显示发送过程
    # server.set_debuglevel(1)
    # 登陆验证 发送邮件 退出
    server.login(from_addr, password)
    server.sendmail(from_addr, [to_addr], msg.as_string())
    server.quit()
# send()

if __name__ == '__main__': 
    msg = sys.argv[1]
    send(msg)



# Adopted from EquiBind
# https://github.com/HannesStark/EquiBind/blob/main/datasets/pdbbind.py

import sys
sys.path.append('../')

import os
import yaml
import shutil
import pickle
import warnings
import os.path as osp
from tqdm import tqdm
from pathlib import Path
from copy import deepcopy

import math
import numpy as np
import pandas as pd
from scipy import spatial
import torch
from torch_geometric.nn import knn_graph
from torch_geometric.transforms import BaseTransform
from torch_geometric.data import Data, InMemoryDataset

import pint
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem.rdPartialCharges import ComputeGasteigerCharges
from Bio.PDB import PDBParser
from Bio.PDB.PDBExceptions import PDBConstructionWarning

from ..utils import pmap_multi, disable_rdkit_logging, safe_index, log, allowable_features, sr
from ..utils import download_url, extract_zip, decide_download


biopython_parser = PDBParser()
ureg = pint.UnitRegistry()
disable_rdkit_logging()


def binary_affinity(affinity, thres=100):
    aff_nM = (affinity * ureg.M).to(ureg.nM).magnitude
    return (aff_nM < thres).float()


class  GenContact(BaseTransform):
    def __init__(self, data_config):
        self.bin_thres = data_config['bin_thres']

    def __call__(self, data):
        edge_index = knn_graph(data.pos, k=5, flow='target_to_source', loop=True, batch=data.batch)
        data.edge_index = edge_index
        data.y = binary_affinity(data.affinity, thres=self.bin_thres)
        if data.y.item() == 0:
            return data

        contact_idx = torch.nonzero(data.contact_label == 1).reshape(-1)
        data.node_label[contact_idx] = 1
        return data


class PLBind(InMemoryDataset):
    def __init__(self, root, data_config, transform, n_jobs=32, debug=False, device='cpu'):
        self.url_raw = 'https://zenodo.org/record/7265547/files/plbind_raw.zip'
        self.url_processed = 'https://zenodo.org/record/7265547/files/plbind_processed.zip'
        self.data_dir = root
        self.use_rdkit_coords = data_config['use_rdkit_coords']
        self.pocket_cutoff = data_config['pocket_cutoff']
        self.bin_thres = data_config['bin_thres']
        self.n_jobs = n_jobs
        self.debug = debug
        from functools import reduce
        def chain_functions(*funcs):
            return lambda x: reduce(lambda v, func: func(v), funcs, x)
        transform = chain_functions(GenContact(data_config), transform)
        super().__init__(root, transform=transform)
        self.data, self.slices, self.idx_split = torch.load(self.processed_paths[0])
        self.data = self.data.to(device)
        self.complex_names = pickle.load(open(os.path.join(self.processed_dir, 'raw_data.pkl'), 'rb'))[-1]
        self.signal_class = 1
        self.dataset_name = 'plbind'
        self.feature_type = data_config['feature_type']

        self.x_dim = self.data.x.shape[1]
        self.pos_dim = self.data.pos.shape[1]

        self.x_lig_dim = self.data.x_lig.shape[1]
        self.pos_lig_dim = self.data.pos_lig.shape[1]

        node_categorical_feat = [len(allowable_features['possible_amino_acids'])]
        node_categorical_feat_lig = [len(allowable_features['possible_atomic_num_list']), len(allowable_features['possible_chirality_list']), len(allowable_features['possible_degree_list']),
                                     len(allowable_features['possible_formal_charge_list']), len(allowable_features['possible_implicit_valence_list']), len(allowable_features['possible_numH_list']),
                                     len(allowable_features['possible_number_radical_e_list']), len(allowable_features['possible_hybridization_list']), len(allowable_features['possible_is_aromatic_list']),
                                     len(allowable_features['possible_numring_list']), len(allowable_features['possible_is_in_ring3_list']), len(allowable_features['possible_is_in_ring4_list']),
                                     len(allowable_features['possible_is_in_ring5_list']), len(allowable_features['possible_is_in_ring6_list']), len(allowable_features['possible_is_in_ring7_list']),
                                     len(allowable_features['possible_is_in_ring8_list'])]

        if self.feature_type == 'only_pos':
            node_scalar_feat_lig = self.pos_lig_dim
            node_scalar_feat = self.pos_dim
            node_categorical_feat, node_categorical_feat_lig = [], []
        elif self.feature_type == 'only_x':
            node_scalar_feat_lig = self.x_lig_dim - 16  # 16 categorical features
            node_scalar_feat = self.x_dim - 1
        elif self.feature_type == 'only_ones':
            node_scalar_feat_lig = 1
            node_scalar_feat = 1
            node_categorical_feat, node_categorical_feat_lig = [], []
        else:
            assert self.feature_type == 'both_x_pos'
            node_scalar_feat_lig = self.x_lig_dim - 16 + self.pos_lig_dim
            node_scalar_feat = self.x_dim - 1 + self.pos_dim

        self.feat_info = {'node_categorical_feat': node_categorical_feat, 'node_scalar_feat': node_scalar_feat}
        self.feat_info_lig = {'node_categorical_feat': node_categorical_feat_lig, 'node_scalar_feat': node_scalar_feat_lig}
        self.n_categorical_feat_to_use_lig = data_config['n_categorical_feat_to_use_lig']
        self.n_scalar_feat_to_use_lig = data_config['n_scalar_feat_to_use_lig']
        self.n_categorical_feat_to_use = data_config['n_categorical_feat_to_use']
        self.n_scalar_feat_to_use = data_config['n_scalar_feat_to_use']
        self.use_lig_info = data_config['use_lig_info']

    @property
    def raw_file_names(self):
        return ['contacts', 'index', 'pdb', 'split']

    @property
    def processed_file_names(self):
        return ['data.pt']

    def download(self):
        if osp.exists(self.processed_paths[0]):
            return
        if decide_download(self.url_raw, is_raw=True):
            path = download_url(self.url_raw, self.root)
            extract_zip(path, self.root)
            os.unlink(path)
        else:
            if decide_download(self.url_processed, is_raw=False):
                path = download_url(self.url_processed, self.root)
                extract_zip(path, self.root)
                os.unlink(path)
            else:
                print('Stop downloading.')
                shutil.rmtree(self.root)
                exit(-1)

    def process(self):
        if decide_download(self.url_processed, is_raw=False):
            path = download_url(self.url_processed, self.root)
            extract_zip(path, self.root)
            os.unlink(path)
            return

        self.contacts_dir = self.data_dir / 'raw' / 'contacts'
        self.index_path = self.data_dir / 'raw' / 'index' / 'INDEX_general_PL_data.2020'
        col_names = ['PDB code', 'resolution', 'release year', '-logKd/Ki', 'Kd/Ki', 'ligand name']
        self.index = pd.read_csv(self.index_path, sep='\s\s+', skiprows=list(range(6)), header=None, names=col_names, engine='python')

        complex_names = get_clean_complex_names(Path(self.raw_dir))
        useable_complex_names = deepcopy(complex_names)
        if not os.path.exists(os.path.join(self.processed_dir, 'raw_data.pkl')):
            ligs, rec_paths, affinities = [], [], []
            for name in tqdm(complex_names, desc='Reading raw data'):
                affinity = self.unit_check(self.index[self.index['PDB code'] == name]['Kd/Ki'].item())
                if affinity is False:
                    useable_complex_names.remove(name)
                    continue

                lig_path = Path(self.raw_dir) / 'pdb' / name
                lig = read_molecule(str(lig_path / f'{name}_ligand.sdf'), sanitize=True, remove_hs=False)
                if lig == None:  # read mol2 file if sdf file cannot be sanitized
                    lig = read_molecule(str(lig_path / f'{name}_ligand.mol2'), sanitize=True, remove_hs=False)

                ligs.append(lig)
                affinities.append(torch.tensor(affinity, dtype=torch.float).view(1, 1))
                rec_paths.append(lig_path / f'{name}_protein_processed.pdb')
            complex_names = useable_complex_names
            log('Saving raw data...')
            pickle.dump((ligs, rec_paths, affinities, complex_names), open(os.path.join(self.processed_dir, 'raw_data.pkl'), 'wb'))
        else:
            log('Loading raw data...')
            ligs, rec_paths, affinities, complex_names = pickle.load(open(os.path.join(self.processed_dir, 'raw_data.pkl'), 'rb'))
        log(f'{len(complex_names)} complexes are usable.')

        if self.debug:
            rec_paths, ligs, complex_names = rec_paths[:100], ligs[:100], complex_names[:100]

        if not os.path.exists(os.path.join(self.processed_dir, 'recs_coors.pkl')):
            receptor_representatives = pmap_multi(get_receptor, zip(rec_paths, ligs), n_jobs=1, cutoff=10, desc='Get receptors')
            recs, recs_coords, c_alpha_coords, n_coords, c_coords, res_nos, chain_ids = map(list, zip(*receptor_representatives))
            log('Saving receptor coordinates...')
            pickle.dump((recs, recs_coords, c_alpha_coords, n_coords, c_coords, res_nos, chain_ids), open(os.path.join(self.processed_dir, 'recs_coors.pkl'), 'wb'))
        else:
            log('Loading receptor coordinates...')
            recs, recs_coords, c_alpha_coords, n_coords, c_coords, res_nos, chain_ids = pickle.load(open(os.path.join(self.processed_dir, 'recs_coors.pkl'), 'rb'))

        if not os.path.exists(os.path.join(self.processed_dir, 'rec_graphs.pkl')):
            rec_graphs = pmap_multi(get_calpha_data, zip(recs, c_alpha_coords), n_jobs=self.n_jobs, desc='Convert receptors to graphs')
            log('Saving receptor graphs...')
            pickle.dump(rec_graphs, open(os.path.join(self.processed_dir, 'rec_graphs.pkl'), 'wb'))
        else:
            log('Loading receptor graphs...')
            rec_graphs = pickle.load(open(os.path.join(self.processed_dir, 'rec_graphs.pkl'), 'rb'))

        if not os.path.exists(os.path.join(self.processed_dir, 'lig_graphs.pkl')):
            lig_graphs = pmap_multi(get_lig_data, zip(ligs, complex_names), n_jobs=1, use_rdkit_coords=self.use_rdkit_coords, debug=self.debug, desc='Convert ligands to graphs')
            log('Saving ligand graphs...')
            pickle.dump(lig_graphs, open(os.path.join(self.processed_dir, 'lig_graphs.pkl'), 'wb'))
        else:
            log('Loading ligand graphs...')
            lig_graphs = pickle.load(open(os.path.join(self.processed_dir, 'lig_graphs.pkl'), 'rb'))

        if not os.path.exists(os.path.join(self.processed_dir, 'pockets.pkl')):
            pockets = pmap_multi(get_pocket_nodes, zip(lig_graphs, rec_graphs, res_nos, chain_ids, complex_names), n_jobs=1, cutoff=self.pocket_cutoff, contacts_dir=self.contacts_dir, desc='Get pockets')
            pocket_nodes, pocket_contacts = map(list, zip(*pockets))
            log('Saving pocket nodes...')
            pickle.dump((pocket_nodes, pocket_contacts), open(os.path.join(self.processed_dir, 'pockets.pkl'), 'wb'))
        else:
            log('Loading pocket nodes...')
            pocket_nodes, pocket_contacts = pickle.load(open(os.path.join(self.processed_dir, 'pockets.pkl'), 'rb'))

        data_list = []
        for idx, name in tqdm(enumerate(complex_names), desc='Processing data'):
            lig_graph = lig_graphs[idx]
            rec_graph = rec_graphs[idx]
            affinity = affinities[idx]

            node_label = pocket_nodes[idx]
            contact_label = pocket_contacts[idx]

            rec_graph.pos = (rec_graph.pos - rec_graph.pos.mean(dim=0, keepdim=True))  # center the graph
            lig_graph.pos = (lig_graph.pos - lig_graph.pos.mean(dim=0, keepdim=True))  # center the graph

            y = binary_affinity(affinity, thres=self.bin_thres)
            data = Data(x=rec_graph.x, pos=rec_graph.pos, true_pos=rec_graph.true_pos,
                        x_lig=lig_graph.x, pos_lig=lig_graph.pos, true_pos_lig=lig_graph.true_pos,
                        affinity=affinity, y=y, node_label=node_label, contact_label=contact_label, pocket_label=node_label)
            data_list.append(data)

        idx_split = self.get_idx_split(self.data_dir, complex_names)
        data, slices = self.collate(data_list)
        torch.save((data, slices, idx_split), self.processed_paths[0])

    def unit_check(self, affinity):
        if 'IC' in affinity:
            return False
        elif '>' in affinity or '<' in affinity:
            return False
        elif '~' in affinity:
            aff_value = affinity.split('~')[-1]
        elif '=' in affinity:
            aff_value = affinity.split('=')[-1]
        else:
            raise ValueError(f'Affinity {affinity} is not in the correct format.')

        unit = aff_value.split('//')[0][-2:]
        aff_value = float(aff_value.split('//')[0][:-2])
        return (aff_value * ureg[unit]).to(ureg['M']).magnitude

    @staticmethod
    def get_idx_split(data_dir, complex_names):
        train_names = (data_dir / 'raw' / 'split' / 'timesplit_no_lig_overlap_train').open().read().splitlines()
        valid_names = (data_dir / 'raw' / 'split' / 'timesplit_no_lig_overlap_val').open().read().splitlines()
        test_names = (data_dir / 'raw' / 'split' / 'timesplit_test').open().read().splitlines()

        train_idx, valid_idx, test_idx, unused_idx = [], [], [], []
        for name in complex_names:
            if name in train_names:
                train_idx.append(complex_names.index(name))
            elif name in valid_names:
                valid_idx.append(complex_names.index(name))
            elif name in test_names:
                test_idx.append(complex_names.index(name))
            else:
                unused_idx.append(complex_names.index(name))
        return {'train': train_idx, 'valid': valid_idx, 'test': test_idx, 'unused': unused_idx}


def get_clean_complex_names(raw_dir):
    complex_names = os.listdir(raw_dir / 'pdb')
    complex_names = [each for each in complex_names if os.listdir(raw_dir / 'pdb' / each) != []]
    assert '1a50' not in complex_names  # no files
    for each in ['3m1s', '3q4c']:  # cannot pickle
        if each in complex_names:
            complex_names.remove(each)
    return complex_names


def read_molecule(molecule_file, sanitize=False, calc_charges=False, remove_hs=False):
    if molecule_file.endswith('.mol2'):
        mol = Chem.MolFromMol2File(molecule_file, sanitize=False, removeHs=False)
    elif molecule_file.endswith('.sdf'):
        supplier = Chem.SDMolSupplier(molecule_file, sanitize=False, removeHs=False)
        mol = supplier[0]
    elif molecule_file.endswith('.pdbqt'):
        with open(molecule_file) as file:
            pdbqt_data = file.readlines()
        pdb_block = ''
        for line in pdbqt_data:
            pdb_block += '{}\n'.format(line[:66])
        mol = Chem.MolFromPDBBlock(pdb_block, sanitize=False, removeHs=False)
    elif molecule_file.endswith('.pdb'):
        mol = Chem.MolFromPDBFile(molecule_file, sanitize=False, removeHs=False)
    else:
        return ValueError('Expect the format of the molecule_file to be '
                          'one of .mol2, .sdf, .pdbqt and .pdb, got {}'.format(molecule_file))

    try:
        if sanitize or calc_charges:
            Chem.SanitizeMol(mol)

        if calc_charges:
            # Compute Gasteiger charges on the molecule.
            try:
                AllChem.ComputeGasteigerCharges(mol)
            except:
                warnings.warn('Unable to compute charges for the molecule.')

        if remove_hs:
            mol = Chem.RemoveHs(mol, sanitize=sanitize)
    except:
        return None
    return mol


def get_pocket_nodes(lig_graph, rec_graph, res_no, chain_id, complex_name, contacts_dir, cutoff):
    node_label = torch.zeros(rec_graph.num_nodes).float()
    lig_rec_distance = spatial.distance.cdist(lig_graph.true_pos, rec_graph.true_pos)
    signal_idx = np.argwhere(lig_rec_distance.min(0) < cutoff).reshape(-1)
    node_label[signal_idx] = 1

    contact_label = torch.zeros(rec_graph.num_nodes).float()
    res_info = np.array(list(zip(res_no, chain_id)))
    contacts = pickle.load(open(contacts_dir / (complex_name + '.pkl'), 'rb'))
    df = pd.DataFrame([], columns=['res_no_rec', 'chain_rec'])
    for each in contacts:
        if each['hydrogen_df'] is not None:
            df = pd.concat([df, each['hydrogen_df'][['res_no_rec', 'chain_rec']].drop_duplicates()], ignore_index=True)
        if each['non_bond_df'] is not None:
            df = pd.concat([df, each['non_bond_df'][['res_no_rec', 'chain_rec']].drop_duplicates()], ignore_index=True)
    df['res_no_rec'] = [int(each) if str(each).lstrip('-').isdigit() else int(each[:-1]) for each in df['res_no_rec']]
    contact_res = df.drop_duplicates().values
    contact_idx = []
    for each_contact in contact_res:
        res = np.where((res_info[:, 0] == str(each_contact[0])) & (res_info[:, 1] == each_contact[1]))[0]
        contact_idx.extend(list(res))
    contact_label[contact_idx] = 1
    return node_label, contact_label


def get_receptor(rec_path, lig, cutoff):
    conf = lig.GetConformer()
    lig_coords = conf.GetPositions()
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=PDBConstructionWarning)
        structure = biopython_parser.get_structure('random_id', rec_path)
        rec = structure[0]  # use the first model

    min_distances, coords, c_alpha_coords, res_nos, chain_ids, n_coords, c_coords, valid_chain_ids, lengths = ([] for i in range(9))
    for i, chain in enumerate(rec):
        chain_coords, chain_c_alpha_coords, chain_n_coords, chain_c_coords, chain_res_nos, chain_chain_ids, chain_is_water, invalid_res_ids = ([] for i in range(8))
        chain_is_water = False
        count = 0
        for _, residue in enumerate(chain):
            if residue.get_resname() == 'HOH':
                chain_is_water = True
            residue_coords = []
            c_alpha, n, c = None, None, None
            res_no = residue.get_id()[1]
            assert isinstance(res_no, int)
            for atom in residue:
                if atom.name == 'CA':
                    c_alpha = list(atom.get_vector())
                if atom.name == 'N':
                    n = list(atom.get_vector())
                if atom.name == 'C':
                    c = list(atom.get_vector())
                residue_coords.append(list(atom.get_vector()))
            # TODO: Also include the chain_coords.append(np.array(residue_coords)) for non amino acids such that they can be used when using the atom representation of the receptor
            if c_alpha != None and n != None and c != None:  # only append residue if it is an amino acid and not some weired molecule that is part of the complex
                chain_c_alpha_coords.append(c_alpha), chain_n_coords.append(n), chain_c_coords.append(c)
                chain_coords.append(np.array(residue_coords)), chain_res_nos.append(res_no), chain_chain_ids.append(chain.get_id())
                count += 1
            else:
                invalid_res_ids.append(residue.get_id())
        for res_id in invalid_res_ids:
            chain.detach_child(res_id)
        if len(chain_coords) > 0:
            all_chain_coords = np.concatenate(chain_coords, axis=0)
            distances = spatial.distance.cdist(lig_coords, all_chain_coords)
            min_distance = distances.min()
        else:
            min_distance = np.inf
        if chain_is_water:
            min_distances.append(np.inf)
        else:
            min_distances.append(min_distance)
        lengths.append(count), coords.append(chain_coords), c_alpha_coords.append(np.array(chain_c_alpha_coords))
        res_nos.append(np.array(chain_res_nos)), chain_ids.append(np.array(chain_chain_ids))
        n_coords.append(np.array(chain_n_coords)), c_coords.append(np.array(chain_c_coords))
        if min_distance < cutoff and not chain_is_water:
            valid_chain_ids.append(chain.get_id())
    min_distances = np.array(min_distances)
    if len(valid_chain_ids) == 0:
        valid_chain_ids.append(np.argmin(min_distances))

    valid_coords, valid_c_alpha_coords, valid_n_coords, valid_c_coords, valid_res_nos, valid_chain_ids_each_res, valid_lengths, invalid_chain_ids = ([] for i in range(8))
    for i, chain in enumerate(rec):  # only use one chain that is closest to the ligand
        if chain.get_id() in valid_chain_ids:
            valid_coords.append(coords[i]), valid_c_alpha_coords.append(c_alpha_coords[i])
            valid_res_nos.append(res_nos[i]), valid_chain_ids_each_res.append(chain_ids[i])
            valid_n_coords.append(n_coords[i]), valid_c_coords.append(c_coords[i]), valid_lengths.append(lengths[i])
        else:
            invalid_chain_ids.append(chain.get_id())
    coords = [item for sublist in valid_coords for item in sublist]  # list with n_residues arrays: [n_atoms, 3]

    c_alpha_coords = np.concatenate(valid_c_alpha_coords, axis=0)  # [n_residues, 3]
    n_coords = np.concatenate(valid_n_coords, axis=0)  # [n_residues, 3]
    c_coords = np.concatenate(valid_c_coords, axis=0)  # [n_residues, 3]
    res_nos = np.concatenate(valid_res_nos, axis=0)  # [n_residues]
    chain_ids = np.concatenate(valid_chain_ids_each_res, axis=0)  # [n_residues]

    for invalid_id in invalid_chain_ids:
        rec.detach_child(invalid_id)

    assert len(c_alpha_coords) == len(n_coords)
    assert len(c_alpha_coords) == len(res_nos)
    assert len(c_alpha_coords) == len(chain_ids)
    assert len(c_alpha_coords) == len(c_coords)
    assert sum(valid_lengths) == len(c_alpha_coords)
    return rec, coords, c_alpha_coords, n_coords, c_coords, res_nos, chain_ids


def get_calpha_data(rec, c_alpha_coords):
    num_residues = len(c_alpha_coords)
    if num_residues <= 1:
        raise ValueError(f"rec contains only 1 residue!")

    node_attr = rec_residue_featurizer(rec)

    residue_representatives_loc_list = []
    for i, residue in enumerate(rec.get_residues()):
        c_alpha_coord = c_alpha_coords[i]
        residue_representatives_loc_list.append(c_alpha_coord)
    residue_representatives_loc_feat = np.stack(residue_representatives_loc_list, axis=0)  # (N_res, 3)
    pos = torch.from_numpy(residue_representatives_loc_feat.astype(np.float32))
    return Data(x=node_attr, pos=pos, true_pos=pos)


def get_lig_data(mol, name, use_rdkit_coords, debug):
    conf = mol.GetConformer()
    true_lig_coords = conf.GetPositions()
    if use_rdkit_coords:
        try:
            rdkit_coords = get_rdkit_coords(mol, debug).numpy()
            R, t = rigid_transform_Kabsch_3D(rdkit_coords.T, true_lig_coords.T)
            lig_coords = ((R @ (rdkit_coords).T).T + t.squeeze())
            if debug:
                log('kabsch RMSD between rdkit ligand and true ligand is ', np.sqrt(np.sum((lig_coords - true_lig_coords) ** 2, axis=1).mean()).item())
        except Exception as e:
            lig_coords = true_lig_coords
            with open('temp_create_dataset_rdkit_timesplit_no_lig_or_rec_overlap_train.log', 'a') as f:
                f.write('Generating RDKit conformer failed for  \n')
                f.write(name)
                f.write('\n')
                f.write(str(e))
                f.write('\n')
                f.flush()
            log('Generating RDKit conformer failed for  ')
            log(name)
            log(str(e))
    else:
        lig_coords = true_lig_coords

    node_attr = lig_atom_featurizer(mol)
    true_pos = torch.from_numpy(np.array(true_lig_coords).astype(np.float32))
    new_pos = torch.from_numpy(np.array(lig_coords).astype(np.float32)) if use_rdkit_coords else true_pos
    return Data(x=node_attr, pos=new_pos, true_pos=true_pos)


def rec_residue_featurizer(rec):
    feature_list = []
    sr.compute(rec, level="R")
    for residue in rec.get_residues():
        sasa = residue.sasa
        for atom in residue:
            if atom.name == 'CA':
                bfactor = atom.bfactor
        assert not np.isinf(bfactor)
        assert not np.isnan(bfactor)
        assert not np.isinf(sasa)
        assert not np.isnan(sasa)
        feature_list.append([safe_index(allowable_features['possible_amino_acids'], residue.get_resname()),
                             sasa,
                             bfactor])
    return torch.tensor(feature_list, dtype=torch.float32)  # (N_res, 1)


def get_rdkit_coords(mol, debug):
    ps = AllChem.ETKDGv2()
    id = AllChem.EmbedMolecule(mol, ps)
    if id == -1:
        if debug:
            log('rdkit coords could not be generated without using random coords. using random coords now.')
        ps.useRandomCoords = True
        AllChem.EmbedMolecule(mol, ps)
        AllChem.MMFFOptimizeMolecule(mol, confId=0)
    else:
        AllChem.MMFFOptimizeMolecule(mol, confId=0)
    conf = mol.GetConformer()
    lig_coords = conf.GetPositions()
    return torch.tensor(lig_coords, dtype=torch.float32)


# R = 3x3 rotation matrix
# t = 3x1 column vector
# This already takes residue identity into account.
def rigid_transform_Kabsch_3D(A, B):
    assert A.shape[1] == B.shape[1]
    num_rows, num_cols = A.shape
    if num_rows != 3:
        raise Exception(f"matrix A is not 3xN, it is {num_rows}x{num_cols}")
    num_rows, num_cols = B.shape
    if num_rows != 3:
        raise Exception(f"matrix B is not 3xN, it is {num_rows}x{num_cols}")


    # find mean column wise: 3 x 1
    centroid_A = np.mean(A, axis=1, keepdims=True)
    centroid_B = np.mean(B, axis=1, keepdims=True)

    # subtract mean
    Am = A - centroid_A
    Bm = B - centroid_B

    H = Am @ Bm.T

    # find rotation
    U, S, Vt = np.linalg.svd(H)

    R = Vt.T @ U.T

    # special reflection case
    if np.linalg.det(R) < 0:
        # print("det(R) < R, reflection detected!, correcting for it ...")
        SS = np.diag([1.,1.,-1.])
        R = (Vt.T @ SS) @ U.T
    assert math.fabs(np.linalg.det(R) - 1) < 1e-5

    t = -R @ centroid_A + centroid_B
    return R, t


def lig_atom_featurizer(mol):
    ComputeGasteigerCharges(mol)  # they are Nan for 93 molecules in all of PDBbind. We put a 0 in that case.
    ringinfo = mol.GetRingInfo()
    atom_features_list = []
    for idx, atom in enumerate(mol.GetAtoms()):
        g_charge = atom.GetDoubleProp('_GasteigerCharge')
        atom_features_list.append([
            safe_index(allowable_features['possible_atomic_num_list'], atom.GetAtomicNum()),
            allowable_features['possible_chirality_list'].index(str(atom.GetChiralTag())),
            safe_index(allowable_features['possible_degree_list'], atom.GetTotalDegree()),
            safe_index(allowable_features['possible_formal_charge_list'], atom.GetFormalCharge()),
            safe_index(allowable_features['possible_implicit_valence_list'], atom.GetImplicitValence()),
            safe_index(allowable_features['possible_numH_list'], atom.GetTotalNumHs()),
            safe_index(allowable_features['possible_number_radical_e_list'], atom.GetNumRadicalElectrons()),
            safe_index(allowable_features['possible_hybridization_list'], str(atom.GetHybridization())),
            allowable_features['possible_is_aromatic_list'].index(atom.GetIsAromatic()),
            safe_index(allowable_features['possible_numring_list'], ringinfo.NumAtomRings(idx)),
            allowable_features['possible_is_in_ring3_list'].index(ringinfo.IsAtomInRingOfSize(idx, 3)),
            allowable_features['possible_is_in_ring4_list'].index(ringinfo.IsAtomInRingOfSize(idx, 4)),
            allowable_features['possible_is_in_ring5_list'].index(ringinfo.IsAtomInRingOfSize(idx, 5)),
            allowable_features['possible_is_in_ring6_list'].index(ringinfo.IsAtomInRingOfSize(idx, 6)),
            allowable_features['possible_is_in_ring7_list'].index(ringinfo.IsAtomInRingOfSize(idx, 7)),
            allowable_features['possible_is_in_ring8_list'].index(ringinfo.IsAtomInRingOfSize(idx, 8)),
            g_charge if not np.isnan(g_charge) and not np.isinf(g_charge) else 0.
        ])

    return torch.tensor(atom_features_list)


if __name__ == '__main__':
    data_dir = Path('../../data/plbind')
    data_config = yaml.safe_load(open('../configs/plbind.yml'))['data']
    dataset = PLBind(data_dir, data_config, n_jobs=32, debug=False)



import torch
import torch.nn as nn
from torch_scatter import scatter
import numpy as np
from ..base import BaseRandom


class LRIBern(BaseRandom):

    def __init__(self, clf, extractor, criterion, config):
        super().__init__()

        self.name = 'lri_bern'
        self.clf = clf
        self.extractor = extractor
        self.criterion = criterion
        self.device = next(self.parameters()).device

        self.pred_loss_coef = config['pred_loss_coef']
        self.info_loss_coef = config['info_loss_coef']
        self.temperature = config['temperature']

        self.final_r = config['final_r']
        self.decay_interval = config['decay_interval']
        self.decay_r = config['decay_r']
        self.init_r = config['init_r']

        self.attn_constraint = config['attn_constraint']

    def __loss__(self, attn, clf_logits, clf_labels, epoch):
        pred_loss = self.criterion(clf_logits, clf_labels.float())

        r = self.get_r(epoch)
        info_loss = (attn * torch.log(attn/r + 1e-6) + (1 - attn) * torch.log((1 - attn)/(1 - r + 1e-6) + 1e-6)).mean()

        pred_loss = self.pred_loss_coef * pred_loss
        info_loss = self.info_loss_coef * info_loss

        loss = pred_loss + info_loss
        loss_dict = {'loss': loss.item(), 'pred': pred_loss.item(), 'info': info_loss.item(), 'r': r}
        return loss, loss_dict

    def forward(self, data, edge_attn=None):
        emb, edge_index = self.clf.get_emb(data, edge_attn=edge_attn)
        original_clf_logits = self.clf.get_pred_from_emb(emb, data.batch)

        node_attn_log_logits = self.extractor(emb, batch=data.batch, pool_out_lig=None)
        if self.attn_constraint == 'smooth_min':
            node_attn_log_logits = scatter(node_attn_log_logits[edge_index[1]].reshape(-1), edge_index[0], reduce='min').reshape(-1, 1)
            node_attn_log_logits = scatter(node_attn_log_logits[edge_index[1]].reshape(-1), edge_index[0], reduce='min').reshape(-1, 1)

        node_attn = self.sampling(node_attn_log_logits, False)
        msk_edge = self.node_attn_to_edge_attn(node_attn, edge_index)
        # edge_attn should be the product of msk_edge and init_edge_attn?
        masked_clf_logits = self.clf(data, edge_attn=msk_edge)
        return masked_clf_logits

    def forward_pass(self, data, epoch, do_sampling):

        emb, edge_index = self.clf.get_emb(data)
        node_attn_log_logits = self.extractor(emb, batch=data.batch, pool_out_lig=None)

        if self.attn_constraint == 'smooth_min': # specially for plbind
            node_attn_log_logits = scatter(node_attn_log_logits[edge_index[1]].reshape(-1), edge_index[0], reduce='min').reshape(-1, 1)
            node_attn_log_logits = scatter(node_attn_log_logits[edge_index[1]].reshape(-1), edge_index[0], reduce='min').reshape(-1, 1)

        node_attn = self.sampling(node_attn_log_logits, do_sampling)


        # node_attn = node_attn * (data.num_nodes / node_attn.sum())
        edge_attn = self.node_attn_to_edge_attn(node_attn, edge_index)
        # masked_clf_logits = self.clf.get_pred_from_emb(node_attn * emb, batch=data.batch)
        masked_clf_logits =  self.clf(data, edge_attn=edge_attn)
        original_clf_logits = self.clf(data)

        loss, loss_dict = self.__loss__(node_attn_log_logits.sigmoid(), masked_clf_logits, data.y, epoch)
        return loss, loss_dict, masked_clf_logits, node_attn.reshape(-1)

    def get_r(self, current_epoch):
        r = self.init_r - current_epoch // self.decay_interval * self.decay_r
        if r < self.final_r:
            r = self.final_r
        return r

    def sampling(self, attn_log_logits, do_sampling):
        if do_sampling:
            random_noise = torch.empty_like(attn_log_logits).uniform_(1e-10, 1 - 1e-10)
            random_noise = torch.log(random_noise) - torch.log(1.0 - random_noise)
            attn_bern = ((attn_log_logits + random_noise) / self.temperature).sigmoid()
        else:
            attn_bern = (attn_log_logits).sigmoid()
        return attn_bern




import argparse
import numpy as np
from nni.experiment import Experiment, ExperimentConfig, AlgorithmConfig, LocalConfig, RemoteConfig, RemoteMachineConfig


def main():
    parser = argparse.ArgumentParser(description='Tuning')
    parser.add_argument('--name', type=str, required=True, help='name of the experiment')
    parser.add_argument('-d', '--dataset', type=str, help='dataset used', required=True)
    parser.add_argument('-m', '--method', type=str, help='method used', required=True)
    parser.add_argument('-b', '--backbone', type=str, help='backbone used', required=True)
    args = parser.parse_args()


    if args.method == 'lri_bern':
        search_space = {
            'epochs': {'_type': 'choice', '_value': [50]},
            'info_loss_coef': {'_type': 'choice', '_value': [0.1, 1.0, 0.01, 10]},
            'final_r': {'_type': 'choice', '_value': [0.5, 0.7, 0.3]},
        }

    elif args.method == 'vgib':
        search_space = {
            'mi_loss_coef': {'_type': 'choice', '_value': [1e-3, 1e-4, 1e-5]},
            'reg_loss_coef': {'_type': 'choice', '_value': [0.1, 1, 10]},
            'noise_loss_coef': {'_type': 'choice', '_value': [10, 1, 0.1]},
        }

    elif args.method == 'ciga':
        search_space = {
            # 'warmup': {'_type': 'choice', '_value': [0, 100, 200]},
            'contrast_loss_coef': {'_type': 'choice', '_value': [0.1, 1, 10]},
            'hinge_loss_coef': {'_type': 'choice', '_value': [0.01, 0.1, 1]},
            'causal_ratio': {'_type': 'choice', '_value': [0.6, 0.7, 0.8]},
        }


    elif args.method == 'lri_gaussian':
        pos_coef_options = {
            'actstrack': [200.0, 300.0],  # [100.0, 200.0, 300.0]  # 1 is ~20%, 10%, 5% percentiles
            'synbind': [10.0, 15.0],  # [5.0, 10.0, 15.0]  # 1 is ~10%, 7%, 5% percentiles
            'tau3mu': [7, 11],  # [3, 7, 11]  # 1 is ~10%, 5%, 3% percentiles
            'plbind': [0.9, 1.2]  # [0.6, 0.9, 1.2]  # 1 is ~10%, 6.5%, 5% percentiles
        }
        kr_options = {
            'actstrack': [10],  # [5, 10]
            'synbind': [5],  # [5, 10]
            'tau3mu': [1.5],  # [1.0, 1.5]
            'plbind': [7]  # [5, 8]
        }

        search_space = {
            'dataset': {'_type': 'choice', '_value': ['acts_2T']},
            'epochs': {'_type': 'choice', '_value': [100]},


            'pred_lr': {'_type': 'choice', '_value': [1.0e-3]},
            'info_loss_coef': {'_type': 'choice', '_value': [0.01, 0.1, 1.0]},

            'pos_coef': {'_type': 'choice', '_value': pos_coef_options[args.dataset]},
            'kr': {'_type': 'choice', '_value': kr_options[args.dataset]},
        }

    elif args.method == 'subgraphx':
        search_space = {
            'subgraph_building_method': {'_type': 'choice', '_value': ['split', 'zero_filling']},
            'high2low': {'_type': 'choice', '_value': [True, False]},
            'score_threshold': {'_type': 'choice', '_value': [0.1, 0.2, 0.3]}
        }

    elif args.method == 'asap':
        search_space = {
            'casual_ratio': {'_type': 'choice', '_value': [0.4, 0.5, 0.6]},
        }

    elif args.method == 'pgmexplainer':
        search_space = {
            'pred_threshold': {'_type': 'choice', '_value': [0, 0.3, 0.5]},
            'percentage': {'_type': 'choice', '_value': [20, 40, 60]},
            'perturb_mode': {'_type': 'choice', '_value': ['split']}
        }

    elif args.method == 'pgexplainer':
        search_space = {
            'epoch': {'_type': 'choice', '_value': [200]},
            # 'dataset': {'_type': 'choice', '_value': ['acts_2T', 'synbind']},
            'size_loss_coef': {'_type': 'choice', '_value': [0.01]},
            'mask_ent_loss_coef': {'_type': 'loguniform', '_value': [0.001, 1]},
        }

    elif args.method == 'gnnexplainer':
        search_space = {
            'size_loss_coef': {'_type': 'choice', '_value': [1.0]},
            'mask_ent_loss_coef': {'_type': 'choice', '_value': [0.01, 0.1]},
            'iter_lr': {'_type': 'uniform', '_value': [0.4, 0.6]}
        }
    else:
        raise ValueError('Unknown method: {}'.format(args.method))


    # if args.dataset == 'all_acts':
    #     assert args.backbone == 'dgcnn'
    #     if args.method == 'psat':
    #         search_space = {
    #             'seed': {'_type': 'choice', '_value': list(range(5))},
    #             'dataset': {'_type': 'choice', '_value': [f'acts_{each}T' for each in range(2, 22, 2)]},
    #             'data/feature_type': {'_type': 'choice', '_value': ['only_pos']},
    #             f'model/{args.backbone}/dropout_p': {'_type': 'choice', '_value': [0.0]},
    #
    #             'psat/warmup': {'_type': 'choice', '_value': [100]},
    #             'psat/epochs': {'_type': 'choice', '_value': [500]},
    #             'psat/pred_lr': {'_type': 'choice', '_value': [1.0e-8]},
    #             'psat/one_gnn': {'_type': 'choice', '_value': [False]},
    #             'psat/info_loss_coef': {'_type': 'choice', '_value': [10.0]},
    #             'psat/pos_coef': {'_type': 'choice', '_value': [100.0]},
    #             'psat/param_u': {'_type': 'choice', '_value': ['sig_dense']},
    #             'psat/dropout_p': {'_type': 'choice', '_value': [0.0]},
    #             'psat/covar_dim': {'_type': 'choice', '_value': [2.0]}
    #         }
    #     else:
    #         assert args.method == 'gradcam'
    #         search_space = {
    #             'seed': {'_type': 'choice', '_value': list(range(5))},
    #             'dataset': {'_type': 'choice', '_value': [f'acts_{each}T' for each in range(2, 22, 2)]},
    #             'data/feature_type': {'_type': 'choice', '_value': ['only_pos']},
    #             f'model/{args.backbone}/dropout_p': {'_type': 'choice', '_value': [0.0]},
    #
    #             'gradcam/warmup': {'_type': 'choice', '_value': [100]},
    #             'gradcam/epochs': {'_type': 'choice', '_value': [1]},
    #             # 'gradcam/target_layers': {'_type': 'choice', '_value': ['last']},
    #             # 'gradcam/way_to_sum': {'_type': 'choice', '_value': ['plain']},
    #             'gradcam/grad_pos': {'_type': 'choice', '_value': [True]},
    #             'gradcam/way_to_sum_pos': {'_type': 'choice', '_value': ['norm']},
    #         }

    trial_per_gpu = 2
    gpu_ratio = 1 / trial_per_gpu
    gpu_index = [3, 4, 5]
    command = f'python trainer.py -d {args.dataset} -m {args.method} -b {args.backbone} --cuda 99'
    config = ExperimentConfig(
        debug=True,
        experiment_name=args.name,
        trial_command=command,
        trial_code_directory='.',
        trial_gpu_number=1,
        trial_concurrency=trial_per_gpu*(len(gpu_index)+1),
        search_space=search_space,
        max_trial_number=int(np.prod([len(v['_value']) for k, v in search_space.items()])),
        experiment_working_directory='../nni-experiments',
        assessor=AlgorithmConfig(name='Medianstop'),
        tuner=AlgorithmConfig(name='GridSearch', class_args={"optimize_mode": 'maximize'}),
        # experiment.config.tuner.name = 'SMAC'
        # experiment.config.max_trial_number = 10000
        training_service=LocalConfig(use_active_gpu=True, max_trial_number_per_gpu=trial_per_gpu, gpu_indices=gpu_index)
        # training_service=RemoteConfig(machine_list=[RemoteMachineConfig(use_active_gpu=True, max_trial_number_per_gpu=\
        # trial_per_gpu, gpu_indices=[0, 1, 2, 3, 4], host='130.207.232.51', user='jzhu617',password='024convdreamStf') ])
        )

    experiment = Experiment(config)
    experiment.run(8082)
    input('Press enter to quit')
    experiment.stop()
    # experiment.view()


if __name__ == '__main__':
    main()


import torch
import torch.nn as nn
from torch_geometric.nn import BatchNorm, InstanceNorm, knn_graph, radius_graph, VGAE
from torch_geometric.nn import GATConv, global_mean_pool, global_add_pool, global_max_pool
import torch_geometric.nn as gnn
import torch.nn.functional as F
from torch_geometric.utils.loop import add_self_loops, remove_self_loops
from torch_geometric.nn.conv.gcn_conv import gcn_norm
from typing import Callable, Union, Tuple
from torch import Tensor
from torch_sparse import SparseTensor, fill_diag
from torch_geometric.typing import OptPairTensor, Adj, OptTensor, Size


class GCNConv(gnn.GCNConv):
    def forward(self, x: Tensor, edge_index: Adj, edge_attr: OptTensor = None,
                edge_weight: OptTensor = None) -> Tensor:
        """"""

        if self.normalize:
            if isinstance(edge_index, Tensor):
                cache = self._cached_edge_index
                if cache is None:
                    edge_index, edge_weight = gcn_norm(  # yapf: disable
                        edge_index, edge_weight, x.size(self.node_dim),
                        self.improved, self.add_self_loops, self.flow)
                    if self.cached:
                        self._cached_edge_index = (edge_index, edge_weight)
                else:
                    edge_index, edge_weight = cache[0], cache[1]

            elif isinstance(edge_index, SparseTensor):
                cache = self._cached_adj_t
                if cache is None:
                    edge_index = gcn_norm(  # yapf: disable
                        edge_index, edge_weight, x.size(self.node_dim),
                        self.improved, self.add_self_loops, self.flow)
                    if self.cached:
                        self._cached_adj_t = edge_index
                else:
                    edge_index = cache

        x = self.lin(x)

        # propagate_type: (x: Tensor, edge_weight: OptTensor)
        out = self.propagate(edge_index, x=x, edge_attr=edge_attr, edge_weight=edge_weight,
                             size=None)

        if self.bias is not None:
            out += self.bias

        return out

    def message(self, x_j: Tensor, edge_attr: Tensor, edge_weight: OptTensor = None) -> Tensor:
        return x_j if edge_weight is None else edge_weight.view(-1, 1) * x_j


class GINEConv(gnn.GINEConv):

    def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj, edge_attr: OptTensor = None, edge_attn: OptTensor = None, size: Size = None) -> Tensor:
        """"""
        if isinstance(x, Tensor):
            x: OptPairTensor = (x, x)

        # propagate_type: (x: OptPairTensor, edge_attr: OptTensor)
        out = self.propagate(edge_index, x=x, edge_attr=edge_attr, edge_attn=edge_attn, size=size)

        x_r = x[1]
        if x_r is not None:
            out += (1 + self.eps) * x_r

        return self.nn(out)

    def message(self, x_j: Tensor, edge_attr: Tensor, edge_attn: OptTensor = None) -> Tensor:
        if self.lin is None and x_j.size(-1) != edge_attr.size(-1):
            raise ValueError("Node and edge feature dimensionalities do not "
                             "match. Consider setting the 'edge_dim' "
                             "attribute of 'GINEConv'")

        if self.lin is not None:
            edge_attr = self.lin(edge_attr)
        m = (x_j + edge_attr).relu()

        if edge_attn is not None:
            return m * edge_attn
        else:
            return m


class GINConv(gnn.GINConv):
    def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj, edge_attr: OptTensor = None, edge_attn: OptTensor = None, size: Size = None) -> Tensor:
        """"""
        if isinstance(x, Tensor):
            x: OptPairTensor = (x, x)

        # propagate_type: (x: OptPairTensor)
        out = self.propagate(edge_index, x=x, edge_attn=edge_attn, size=size)

        x_r = x[1]
        if x_r is not None:
            out += (1 + self.eps) * x_r

        return self.nn(out)

    def message(self, x_j: Tensor, edge_attn: OptTensor = None) -> Tensor:
        if edge_attn is not None:
            return x_j * edge_attn
        else:
            return x_j

class MySequential(nn.Sequential):
    def forward(self, *inputs, **kwargs):
        for module in self._modules.values():
            if type(inputs) == tuple:
                inputs = module(*inputs)
            elif isinstance(module, InstanceNorm):
                inputs = module(inputs, batch=kwargs['batch'])
            else:
                inputs = module(inputs)
        return inputs
# class GINConv(gnn.GINConv):
#
#     def __init__(self, nn: Callable, eps: float = 0., train_eps: bool = False,
#                  **kwargs):
#         super().__init__(nn, eps, train_eps, **kwargs)
#         self.edge_weight = None
#         self.fc_steps = None
#         self.reweight = None
#
#     # def children(self):
#     #     if
#     #     return iter([])
#
#
#     def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj,
#                 edge_weight: OptTensor = None, **kwargs) -> Tensor:
#         """"""
#         self.num_nodes = x.shape[0]
#         if isinstance(x, Tensor):
#             x: OptPairTensor = (x, x)
#
#         # propagate_type: (x: OptPairTensor)
#         if edge_weight is not None:
#             self.edge_weight = edge_weight
#             assert edge_weight.shape[0] == edge_index.shape[1]
#             self.reweight = False
#         else:
#             edge_index, _ = remove_self_loops(edge_index)
#             self_loop_edge_index, _ = add_self_loops(edge_index, num_nodes=self.num_nodes)
#             if self_loop_edge_index.shape[1] != edge_index.shape[1]:
#                 edge_index = self_loop_edge_index
#             self.reweight = True
#         out = self.propagate(edge_index, x=x[0], size=None)
#
#         # if data_args.task == 'explain':
#         #     layer_extractor = []
#         #     hooks = []
#         #
#         #     def register_hook(module: nn.Module):
#         #         if not list(module.children()):
#         #             hooks.append(module.register_forward_hook(forward_hook))
#         #
#         #     def forward_hook(module: nn.Module, input: Tuple[Tensor], output: Tensor):
#         #         # input contains x and edge_index
#         #         layer_extractor.append((module, input[0], output))
#         #
#         #     # --- register hooks ---
#         #     self.nn.apply(register_hook)
#         #
#         #     nn_out = self.nn(out)
#         #
#         #     for hook in hooks:
#         #         hook.remove()
#         #
#         #     fc_steps = []
#         #     step = {'input': None, 'module': [], 'output': None}
#         #     for layer in layer_extractor:
#         #         if isinstance(layer[0], nn.Linear):
#         #             if step['module']:
#         #                 fc_steps.append(step)
#         #             # step = {'input': layer[1], 'module': [], 'output': None}
#         #             step = {'input': None, 'module': [], 'output': None}
#         #         step['module'].append(layer[0])
#         #         if kwargs.get('probe'):
#         #             step['output'] = layer[2]
#         #         else:
#         #             step['output'] = None
#         #
#         #     if step['module']:
#         #         fc_steps.append(step)
#         #     self.fc_steps = fc_steps
#         # else:
#         nn_out = self.nn(out)
#
#         return nn_out
#
#     def message(self, x_j: Tensor) -> Tensor:
#         if self.reweight:
#             edge_weight = torch.ones(x_j.shape[0], device=x_j.device)
#             edge_weight.data[-self.num_nodes:] += self.eps
#             edge_weight = edge_weight.detach().clone()
#             edge_weight.requires_grad_(True)
#             self.edge_weight = edge_weight
#         return x_j * self.edge_weight.view(-1, 1)



class ExtractorMLP(nn.Module):

    def __init__(self, config):
        super().__init__()
        self.norm_type = config.norm_type
        dropout_p = config.dropout_p
        act_type = config.act_type
        # level = getattr(config, 'x_level', None)
        hidden_size = config.hidden_size

        self.pos_dim = getattr(config, "pos_dim", None)
        self.edge_extractor = MLP([hidden_size * 2, hidden_size * 4, hidden_size * 2, 1], dropout_p, self.norm_type, act_type)
        self.node_extractor = MLP([hidden_size, hidden_size * 2, hidden_size, 1], dropout_p, self.norm_type, act_type)
        if self.pos_dim:
            self.gaussian_extractor = MLP([hidden_size, hidden_size * 2, hidden_size, config.pos_dim ** 2 + 2], dropout_p, self.norm_type, act_type)

    def forward(self, node_emb, data, att_type):
        col, row = data.edge_index
        edge_emb = torch.cat([node_emb[row], node_emb[col]], dim=-1)
        emb = edge_emb if att_type == 'edge' else node_emb
        feature_extractor = self.edge_extractor if att_type == 'edge' else self.node_extractor if att_type == 'node' else self.gaussian_extractor
        if self.norm_type == 'instance':
            batch = data.batch if emb.shape[0] == data.num_nodes else data.batch[data.edge_index[0]]
            attn_log_logits = feature_extractor(emb, batch=batch)
        else:
            attn_log_logits = feature_extractor(emb)
        return attn_log_logits

class VGAE3MLP(VGAE):
    def __init__(self, config, dataset):
        dropout_p = config.dropout_p
        self.encode_size = config.encode_size #16
        decode_size = config.decode_size #32
        init_size = dataset.num_features
        self.output_size = dataset.num_classes
        encoder = GCNEncoder(init_size, self.encode_size, dropout_p, dataset)
        decoder = InnerProductDecoderMLP(self.encode_size, decode_size, dropout_p)
        super(VGAE3MLP, self).__init__(encoder, decoder)
        # self.dc = InnerProductDecoderMLP(output_dim, decoder_hidden_dim1, decoder_hidden_dim2, dropout, act=lambda x: x)

    def vgae_loss(self, recoverd_adj, org_adj, mu, logvar, data):
        n_nodes = org_adj.shape[0]
        adj = org_adj - torch.eye(n_nodes)
        pos_weight = (adj.shape[0] * adj.shape[0] - adj.sum()) / adj.sum()
        norm =adj.shape[0] * adj.shape[0] / ((adj.shape[0] * adj.shape[0] - adj.sum()) * 2)
        bce = F.binary_cross_entropy_with_logits(recoverd_adj.flatten(1).T, org_adj.flatten(1).T, pos_weight=pos_weight,
                                                 reduction='none').mean()
        cost = norm * bce
        KLD = -0.5 / n_nodes * torch.mean(torch.sum(
            1 + 2 * logvar - mu.pow(2) - logvar.exp().pow(2), -1))
        return cost

class MLP(MySequential):
    def __init__(self, channels, dropout_p, norm_type, act_type):
        norm = self.get_norm(norm_type)
        act = self.get_act(act_type)

        m = []
        for i in range(1, len(channels)):
            m.append(nn.Linear(channels[i - 1], channels[i]))

            if i < len(channels) - 1:
                m.append(norm(channels[i]))
                m.append(act())
                m.append(nn.Dropout(dropout_p))

        super(MLP, self).__init__(*m)

    @staticmethod
    def get_norm(norm_type):
        if isinstance(norm_type, str) and 'batch' in norm_type:
            return BatchNorm
        elif norm_type == 'none' or norm_type is None:
            return nn.Identity
        elif norm_type == 'instance':
            return InstanceNorm
        else:
            raise ValueError('Invalid normalization type: {}'.format(norm_type))

    @staticmethod
    def get_act(act_type):
        if act_type == 'relu':
            return nn.ReLU
        elif act_type == 'silu':
            return nn.SiLU
        elif act_type == 'none':
            return nn.Identity
        else:
            raise ValueError('Invalid activation type: {}'.format(act_type))

class GCNEncoder(nn.Module):
    def __init__(self, input_size, hidden_size, dropout, dataset):
        super(GCNEncoder, self).__init__()
        feat_info = getattr(dataset, "feat_info", None)
        if feat_info['node_categorical_feat']:
            self.emb = FeatEncoder(hidden_size, feat_info['node_categorical_feat'], feat_info['node_scalar_feat'])
            input_size = hidden_size
            assert dataset.name != 'plbind'
        self.gc1 = MySequential(GCNConv(input_size, hidden_size, dropout=dropout), nn.ReLU())
        self.gc1_1 = MySequential(GCNConv(hidden_size, hidden_size, dropout=dropout), nn.ReLU())
        self.gc2 = MySequential(GCNConv(hidden_size, hidden_size, dropout=dropout), nn.Identity())
        self.gc3 = MySequential(GCNConv(hidden_size, hidden_size, dropout=dropout), nn.Identity())

    def forward(self, x, edge_index):
        if getattr(self, "emb", None):
            x = self.emb(x)
        hidden1 = self.gc1(x, edge_index)
        hidden2 = self.gc1_1(hidden1, edge_index)
        return self.gc2(hidden2, edge_index), self.gc3(hidden2, edge_index)

class InnerProductDecoderMLP(nn.Module):
    """Decoder for using inner product for prediction."""
    def __init__(self, input_dim, hidden_dim, dropout):
        super(InnerProductDecoderMLP, self).__init__()
        self.fc = nn.Linear(input_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim)
        self.dropout = dropout
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.fc.weight)
        torch.nn.init.zeros_(self.fc.bias)
        torch.nn.init.xavier_uniform_(self.fc2.weight)
        torch.nn.init.zeros_(self.fc2.bias)

    def forward(self, z, sigmoid=True):
        z = F.relu(self.fc(z))
        z = torch.sigmoid(self.fc2(z))
        z = F.dropout(z, self.dropout, training=self.training)
        if len(z.shape) == 3:
            adj = torch.bmm(z, torch.transpose(z, 1, 2))
        else:
            adj = torch.matmul(z, z.t())

        return torch.sigmoid(adj) if sigmoid else adj

class CoorsNorm(nn.Module):
    def __init__(self, eps = 1e-6, scale_init = 1.):
        super().__init__()
        self.eps = eps
        # scale = torch.zeros(1).fill_(scale_init)
        # self.scale = nn.Parameter(scale)

    def forward(self, coors):
        norm = coors.norm(dim = -1, keepdim = True)
        normed_coors = coors / norm.clamp(min = self.eps)
        return normed_coors #* self.scale


class FeatEncoder(torch.nn.Module):

    def __init__(self, hidden_size, categorical_feat, scalar_feat, n_categorical_feat_to_use=-1, n_scalar_feat_to_use=-1):
        super().__init__()
        self.embedding_list = torch.nn.ModuleList()

        self.num_categorical_feat = len(categorical_feat)
        self.n_categorical_feat_to_use = self.num_categorical_feat if n_categorical_feat_to_use == -1 else n_categorical_feat_to_use
        self.num_scalar_feat_to_use = scalar_feat if n_scalar_feat_to_use == -1 else n_scalar_feat_to_use

        for i in range(self.n_categorical_feat_to_use):
            num_categories = categorical_feat[i]
            emb = torch.nn.Embedding(num_categories, hidden_size)
            self.embedding_list.append(emb)

        if self.num_scalar_feat_to_use > 0:
            assert n_scalar_feat_to_use == -1
            self.linear = torch.nn.Linear(self.num_scalar_feat_to_use, hidden_size)

        total_cate_dim = self.n_categorical_feat_to_use*hidden_size
        self.dim_mapping = torch.nn.Linear(total_cate_dim + hidden_size, hidden_size) if self.num_scalar_feat_to_use > 0 else torch.nn.Linear(total_cate_dim, hidden_size)

    def forward(self, x):
        x_embedding = []
        for i in range(self.n_categorical_feat_to_use):
            x_embedding.append(self.embedding_list[i](x[:, i].long()))

        if self.num_scalar_feat_to_use > 0:
            x_embedding.append(self.linear(x[:, self.num_categorical_feat:]))

        x_embedding = self.dim_mapping(torch.cat(x_embedding, dim=-1))
        return x_embedding
